import json
import logging
from argparse import ArgumentParser

import dtiot_d2c.d2c as d2c
import dtiot_d2c.d2c.test.utils as tutils
import dtiot_d2c.dmo as dmo
from dtiot_d2c.cli.cli_command import ApiConnCmdLineInterface
from dtiot_d2c.d2c import CMDARGS, Application, Device, DeviceGroup, OneM2MBase
from dtiot_d2c.d2c import utils as utils
from dtiot_d2c.d2c.get_response import GetResponse
from dtiot_d2c.dmo import dmoCLI as dmocli

log = logging.getLogger(__name__)
            
#####################################################################################                
class VerifyDeviceGroupConsistencyCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Verifies the consistency of a device group."
    
    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-n",required=True)

    def main(self, cmdargs=None):
        
        device_group_name = cmdargs.name
        api = self.apiConn
        
        device_group:DeviceGroup = DeviceGroup.get(api, name=device_group_name)

        if not device_group:
            raise Exception(f"Could not load device group {device_group_name}.")

        GetResponse(response=device_group).print()      
        
        ###
        # Load referenced applications
        applications = []
        for id in device_group.applicationIds:
            print(f"Loading application {id} ... ", end="", flush=True)
            application:Application = Application.get(api, id=id)

            if not application:
                tutils.print_ERROR(f"ERROR: could not load application.", flush=True)
            else:
                tutils.print_OK(flush=True)
            applications.append(application)            

        ###
        # Load referenced devices
        devices = []
        for id in device_group.deviceIds:
            print(f"Loading device {id} ... ", end="", flush=True)
            device:Device = Device.get(api, id=id)

            if not device:
                tutils.print_ERROR(f"ERROR: could not load device.", flush=True)
            else:
                tutils.print_OK(flush=True)

            devices.append(device)            
        
        ###
        # Check the device communication containers
        for device in devices:
            for application in applications:
                device_name = device.name
                application_name = application.name

                n = f"device-communication/{device_name}/received-msg/{device_group_name}-{application_name}-{device_name}"
                print(f"Checking {n} ... ", end="", flush=True)
                
                onem2m_dict = dmo.get_resources(api, resourceType="subscription", resourceName=n)
                if not onem2m_dict:
                    tutils.print_ERROR(f"ERROR: subscription does not exist.", flush=True)
                else:
                    tutils.print_OK(flush=True)                        

                
        
        


import json
import logging
import os
import re
import time
import uuid
from datetime import datetime
from typing import List

from pydantic import BaseModel, Field, PrivateAttr

import dtiot_d2c.dmo as dmo
from dtiot_d2c.d2c import utils as utils
from dtiot_d2c.d2c.onem2m_base import OneM2MBase
from dtiot_d2c.dmo import ApiClient, device_provisioning

from .consts import *
from .message import Message

log = logging.getLogger(__name__)
class Device(OneM2MBase):
    _init_with_dvi: bool = PrivateAttr(default=False)
    _init_with_crds: bool = PrivateAttr(default=False)
    _dirty_dvi:bool = PrivateAttr(default=False)
    _dirty_credentials:bool = PrivateAttr(default=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        if self.onem2m:
            if self.onem2m.get("m2m:dvi", []):
                self._init_with_dvi = True
            if self.onem2m.get("m2m:crds", []):
                self._init_with_crds = True
        
    @property        
    def dirty_dvi(self)->bool:
        return self._dirty_dvi    
    @dirty_dvi.setter
    def dirty_dvi(self, value:bool):
        self._dirty_dvi = value        

    @property        
    def dirty_credentials(self)->bool:
        return self._dirty_credentials        
    @dirty_credentials.setter
    def dirty_credentials(self, value:bool):
        self._dirty_credentials = value          
        
    @property
    def iccid(self)->List[str]:
        return self._get_label("ICCID")

    @iccid.setter
    def iccid(self, value:str):
        self._set_label("ICCID", value)
    
    ### 
    # Uplink properties
    def set_uplink_properties(self, props:dict):
        self._set_prefixed_labels(UPLINK_PROPERTY_PREFIX, props)

    def delete_uplink_properties(self):
        self._delete_prefixed_labels(UPLINK_PROPERTY_PREFIX)

    def set_uplink_property(self, key, value):
        self._set_prefixed_label(UPLINK_PROPERTY_PREFIX, key, value)
        
    def del_uplink_property(self, key):
        self._del_perfixed_property(UPLINK_PROPERTY_PREFIX, key)

    @property    
    def uplinkProperties(self)->dict:   
        return self._get_prefixed_labels(UPLINK_PROPERTY_PREFIX)
    
    @uplinkProperties.setter
    def uplinkProperties(self, props:dict):
        self.set_uplink_properties(props)    

    ### 
    # Device properties
    def delete_device_properties(self):
        old_labels = self.labels
        new_labels = {}
        for (k, v) in old_labels.items():
            if k.startswith(UPLINK_PROPERTY_PREFIX) or k in DEVICE_SYSTEM_LABELS:
                new_labels[k] = old_labels[k]
        self.labels = new_labels

    def set_device_properties(self, props:dict):
        if props == None:
            return

        self.delete_device_properties()
        
        for (key, value) in props.items():
            self._set_label(key, value)        
            
    def set_device_property(self, key, value):
        self._set_label(key, value)

    def del_device_property(self, key):
        self._del_label(key)

    @property    
    def deviceProperties(self)->dict:
        d = {}
        for (k, v) in self.labels.items():
            if k.startswith(UPLINK_PROPERTY_PREFIX) or k in DEVICE_SYSTEM_LABELS:
                continue
            d[k] = v
        return d
    
    @deviceProperties.setter
    def deviceProperties(self, props:dict):
        self.set_device_properties(props)
    
    ### DVI attributes
        
    @property
    def deviceType(self)->List[str]:
        return self._get_child_array_attribute("m2m:dvi", "dty")

    @deviceType.setter
    def deviceType(self, value:str):
        self._set_child_array_attribute("m2m:dvi", "dty", value)
        
    @property
    def deviceName(self)->List[str]:
        return self._get_child_array_attribute("m2m:dvi", "dvnm")

    @deviceName.setter
    def deviceName(self, value:str):
        self._set_child_array_attribute("m2m:dvi", "dvnm", value)

    @property
    def location(self)->List[str]:
        return self._get_child_array_attribute("m2m:dvi", "loc")

    @location.setter
    def location(self, value:str):
        self._set_child_array_attribute("m2m:dvi", "loc", value)

    @property
    def label(self)->List[str]:
        return self._get_child_array_attribute("m2m:dvi", "dlb")

    @label.setter
    def label(self, value:str):
        self._set_child_array_attribute("m2m:dvi", "dlb", value)

    @property
    def description(self)->List[str]:
        return self._get_child_array_attribute("m2m:dvi", "dc")

    @description.setter
    def description(self, value:str):
        self._set_child_array_attribute("m2m:dvi", "dc", value)
                
    @property
    def firmwareVersion(self)->List[str]:
        return self._get_child_array_attribute("m2m:dvi", "fwv")

    @firmwareVersion.setter
    def firmwareVersion(self, value:str):
        self._set_child_array_attribute("m2m:dvi", "fwv", value)
        
    @property
    def softwareVersion(self)->List[str]:
        return self._get_child_array_attribute("m2m:dvi", "swv")

    @softwareVersion.setter
    def softwareVersion(self, value:str):
        self._set_child_array_attribute("m2m:dvi", "swv", value)
                          
    @property
    def osVersion(self)->List[str]:
        return self._get_child_array_attribute("m2m:dvi", "osv")

    @osVersion.setter
    def osVersion(self, value:str):
        self._set_child_array_attribute("m2m:dvi", "osv", value)

    @property
    def hardwareVersion(self)->List[str]:
        return self._get_child_array_attribute("m2m:dvi", "hwv")

    @hardwareVersion.setter
    def hardwareVersion(self, value:str):
        self._set_child_array_attribute("m2m:dvi", "hwv", value)
                     
    @property
    def protocols(self)->List[str]:#
        return self._get_child_array_attribute("m2m:dvi", "ptl")

    @protocols.setter
    def protocols(self, value:List[str]):
        self._set_child_array_attribute("m2m:dvi", "ptl", value)
        
    @property
    def manufacturer(self)->List[str]:
        return self._get_child_array_attribute("m2m:dvi", "man")

    @manufacturer.setter
    def manufacturer(self, value:str):
        self._set_child_array_attribute("m2m:dvi", "man", value)
                
    @property
    def model(self)->List[str]:
        return self._get_child_array_attribute("m2m:dvi", "mod")

    @model.setter
    def model(self, value:str):
        self._set_child_array_attribute("m2m:dvi", "mod", value)
                        
    @property
    def subModel(self)->List[str]:
        return self._get_child_array_attribute("m2m:dvi", "smod")

    @subModel.setter
    def subModel(self, value:str):
        self._set_child_array_attribute("m2m:dvi", "smod", value)

    @property
    def country(self)->List[str]:
        return self._get_child_array_attribute("m2m:dvi", "cnty")

    @country.setter
    def country(self, value:str):
        self._set_child_array_attribute("m2m:dvi", "cnty", value)

    ### CRDS attributes (Credentials)
    '''
    "m2m:crds": [
        {
            "dc": "cred",
            "mgd": 1029,
            "rn": "credentials",
            "ty": 13,
            "ri": "67fff2fceab987515f7c361c",
            "pi": "67fff2fceab987515f7c3616",
            "ct": "20250416T181212,718000",
            "lt": "20250416T181212,718000",
            "crid": "device01-115",
            "crse": "Q3hhc21FTUw3NFNiUlIzUw=="
        }
    ],    
    '''    
    @property
    def credentialsId(self)->List[str]:
        return self._get_child_array_attribute("m2m:crds", "crid")

    @credentialsId.setter
    def credentialsId(self, value:str):
        self._set_child_array_attribute("m2m:crds", "crid", value)
        
    @property
    def credentialsSecret(self)->List[str]:#
        if os.getenv("SHOW_SECRETS", "false").lower() in ["true", "1", "yes"]:
            return self._get_child_array_attribute("m2m:crds", "crse")
        else:
            return "***"

    @credentialsSecret.setter
    def credentialsSecret(self, value:str):
        self._set_child_array_attribute("m2m:crds", "crse", value)

    def validate_name(name:str)->bool:
        return re.match(VALID_NAME_CHARS, name)

    def model_dump(self,  *args, **kwargs):
        md = super().model_dump(*args, **kwargs)
        for key in ["iccid",
                    "deviceType",
                    "deviceName",
                    "location", 
                    "label",
                    "description",
                    "deviceName",
                    "firmwareVersion", "softwareVersion", "osVersion", "hardwareVersion",
                    "protocols", 
                    "manufacturer", "model", "subModel",
                    "country",
                    "uplinkProperties",
                    "deviceProperties",
                    "credentialsId",
                    "credentialsSecret"
                   ]:
            md[key] = getattr(self, key)
        del md["labels"]

        return md

    def set_from_kwargs(self, **kwargs):
        super().set_from_kwargs(**kwargs)

        #### Copy simple string argumentsw
        for (paramName, attrName) in (("iccid", "iccid"),
                                      ("description", "description"),
                                      ("label", "label"),
                                      ("device_type", "deviceType"),
                                      ("device_name", "deviceName"),
                                      ("firmware_version", "firmwareVersion"),
                                      ("software_version", "softwareVersion"),
                                      ("os_version", "osVersion"),
                                      ("hardware_version", "hardwareVersion"),
                                      ("location", "location"),
                                      ("manufacturer", "manufacturer"),
                                      ("model", "model"),
                                      ("sub_model", "subModel"),
                                      ("country", "country"),
                                      ("credentials_id", "credentialsId"),
                                      ("credentials_secret", "credentialsSecret")
                                     ):
            if not (value := kwargs.get(paramName, None)):
                continue

            setattr(self, attrName, value)
                  
            if paramName in ("credentials_id", "credentials_secret"):
                self.dirty_credentials = True

            elif paramName in ("description", "label", "device_type", "device_name", 
                               "firmware_version", "software_version", "os_version", 
                               "hardware_version", "location", "manufacturer", "model", 
                               "sub_model", "country"):
                self.dirty_dvi = True
            
        ### Copy list args into the object
        for (paramName, attrName) in [["protocols", "protocols"],
                                     ]:
            if not (value := kwargs.get(paramName, None)):
                continue
            if type(value) == list:
                setattr(self, attrName, value)
            elif type(value) == str:
                setattr(self, attrName, utils.splitString(value, ","))
            else:
                log.error(f"Unsupported type {type(value)} for argument {paramName}={value}")                  

        ### 
        # Copy the uplink properties into the labels of the device
        self.integrate_labels(kwargs.get("uplink_properties", None), 
                              self.del_uplink_property,
                              self.set_uplink_property,
                              self.set_uplink_properties)

        ###
        # Copy the ddevice properties into the labels of the device
        self.integrate_labels(kwargs.get("device_properties", None), 
                              self.del_device_property,
                              self.set_device_property,
                              self.set_device_properties)
                
    def to_onem2m(self, **kwargs)->dict:
        return {}

    def to_onem2m_dvi(self)->dict:
        '''
        {
            "m2m:dvi": {
                "loc": "geo:48.861069,2.335798",
                "rn": "device001",
                "mod": "Heart of Gold",
                "man": "a Manufacturer",
                "dlb": "|label:value anotherLabel:value",
                "mgd": 1007,
                "dty": "SIM"
            }
        }
        '''
        dvi = {
            "rn": DEVICE_INFO_RN,
            "mgd": DEVICE_INFO_MGD
        }
        
        for (dev_attr_name, onem2m_path) in [["location",       "loc"],
                                            ["deviceType",      "dty"],
                                            ["deviceName",      "dvnm"],
                                            ["label",           "dlb"],
                                            ["description",     "dc"],
                                            ["firmwareVersion", "fwv"],
                                            ["softwareVersion", "swv"],
                                            ["osVersion",       "osv"],
                                            ["hardwareVersion", "hwv"],
                                            ["protocols",       "ptl"],
                                            ["manufacturer",    "man"],
                                            ["model",           "mod"],
                                            ["subModel",        "smod"],
                                            ["country",         "cnty"],
                                            ]:
            utils.setValueInDict(dvi, onem2m_path, getattr(self, dev_attr_name), createIfNotExist=True)
            
        return dvi

    def to_onem2m_crds(self)->dict:
        '''
        {
            "dc": "cred",
            "mgd": 1029,
            "rn": "credentials",
            "ty": 13,
            "ri": "680be5a783dba99abdafb9ab",
            "pi": "680be5a783dba99abdafb9a5",
            "ct": "20250425T194231,895000",
            "lt": "20250425T194231,895000",
            "crid": "UNDEFINED",
            "crse": "UNDEFINED"
        }
        '''
        crds = {
            "dc":"cred",
            "mgd": 1029,
            "rn": "credentials",
            "crid": self.credentialsId,
            "crse": self.credentialsSecret
        }
            
        return crds

    ######################################################################
    # Static public functions
    ######################################################################
    def get_create_onem2m_dict(name:str, **kwargs)->dict:
        return {}
            
    def get(api_client:ApiClient,  
            origin:str=None,       
            id:str=None,
            name:str=None,
            select:str=None,                 # Comma separated list of element names which shall be selected.
            format:str="CSV",                # Output format. Possible values are CSV and JSON
            resultContentIndocator:str=None, # Result content indicator. Supported types are RCIs (see above).
            limit:int=1000000000,            # Max number of objects to return.
            offset:int=0
           ) -> dict:  

        return get(api_client, origin=origin, id=id, name=name, resultContentIndocator=resultContentIndocator,
                   limit=limit, offset=offset, select=select, format=format)
        
    def create(api_client:ApiClient,
               origin:str=None,     
               name:str=None,
               **kwargs):    
        return create(api_client, origin, name, **kwargs)
    
    def delete(api_client:ApiClient, 
               origin:str=None, 
               id:str=None,
               name:str=None 
              ):
        return delete(api_client, origin=origin, id=id, name=name)
    
    def update(api_client:ApiClient, 
               origin:str=None, 
               device=None, 
               name:str=None,
               id:str=None,
               **kwargs
              ):  
        return update(api_client, 
                      origin, 
                      device=device, 
                      name=name,
                      id=id,
                      **kwargs)  

    def inject_uplink_message(api_client:ApiClient, 
                              origin:str=None,      
                              id:str=None,
                              name:str=None,
                              device=None,
                              content=None,
                              content_type:str="text/plain",
                              contentEncoding:str="none",
                             ):
        return _inject_message(api_client,
                               origin=origin,
                               msg_store=INBOUND_UPLINK_MESSAGE_STORE,
                               id=id,
                               name=name,
                               device=device,
                               content=content,
                               content_type=content_type,
                               contentEncoding=contentEncoding,
                              )
        
    def inject_downlink_message(api_client:ApiClient, 
                                origin:str=None,      
                                id:str=None,
                                name:str=None,
                                device=None,
                                content=None,
                                content_type:str="text/plain",
                                contentEncoding:str="none",
                               ):
        return _inject_message(api_client,
                               origin=origin,
                               msg_store=INBOUND_DOWNLINK_MESSAGE_STORE,
                               id=id,
                               name=name,
                               device=device,
                               content=content,
                               content_type=content_type,
                               contentEncoding=contentEncoding,
                              )        
        
    def get_messages(api_client:ApiClient,  
                     origin:str=None,       
                     name:str=None,
                     id:str=None,
                     msg_store:MessageStore=None,
                     msg_name:str=None,
                     last:bool=False,
                     select:str=None,              
                     format:str="CSV",                # Output format. Possible values are CSV and JSON
                     limit:int=1000000000,  
                     offset:int=0,         
                    ) -> dict:  

        return get_messages(api_client, origin=origin, name=name, id=id, 
                            msg_store=msg_store, msg_name=msg_name, 
                            last=last, limit=limit, offset=offset, 
                            select=select, format=format)
        
    def delete_message(api_client:ApiClient,
                       origin:str=None,   
                       name:str=None,    
                       id:str=None,
                       msg_store:MessageStore=None,
                       msg_name:str=None,
                       msg_id:str=None
                       ):
        delete_message(api_client,
                       origin=origin,
                       name=name,
                       id=id,
                       msg_store=msg_store,
                       msg_name=msg_name,
                       msg_id=msg_id)
                          
################################################################
################################################################
################################################################
#MARK: IMPLEMENTATIONS

def _get(api_client:ApiClient,  
         origin:str=None,   
         id:str=None,    
         name:str=None,
         resultContentIndocator:str=None, # Result content indicator. Supported types are RCIs (see above).
         limit:int=1000000000,            # Max number of objects to return.
         offset:int=0,
         select:str=None,                 # Comma separated list of element names which shall be selected.
         format:str="CSV"                 # Output format. Possible values are CSV and JSON
        ) -> dict:        
    if not name and not id:
        response = dmo.get_resources(api_client, 
                                     origin=origin,
                                     resourceType="nod", 
                                     limit=limit,
                                     offset=offset,
                                     resultContentIndocator=resultContentIndocator)
    else:
        response = dmo.get_resources(api_client, 
                                     origin=origin,
                                     resourceId=id,
                                     resourceName=name,
                                     limit=limit,
                                     offset=offset,
                                     resultContentIndocator=resultContentIndocator)      
    if not response:
        return None

    if not name and not id:
        object = [OneM2MBase(onem2m=onem2m) for onem2m in response]
    else:
        object = Device(onem2m=response)  
                
    # Return the plain objects
    if not select:
        return object
    else:
        return dmo.select_from_response_body(object, select=select, format=format)

def get(api_client:ApiClient,
        origin:str=None,   
        id:str=None,
        name:str=None,    
        select:str=None,                 
        format:str="CSV",             
        limit:int=1000000000,      
        offset:int=0,   
        resultContentIndocator:str=None, 
       ):        
    if name or id:
        log.info(f"Getting device {name or id} from DMO ...")
    else:
        log.info(f"Getting device from DMO ...")        

    return _get(api_client, origin=origin, id=id, name=name, select=select, format=format, 
                resultContentIndocator=resultContentIndocator, limit=limit, offset=offset)

def delete(api_client:ApiClient, 
           origin:str=None, 
           id:str=None,
           name:str=None, 
          ):
    if not name and not id:
        raise Exception(f"delete() requires parameter name or id as input.")
    
    log.info(f"Deleting device {name} from DMO ...")
    return dmo.delete_resource(api_client, origin=origin, name=name, id=id)

def _update(api_client:ApiClient,  # Client connection definition.
            origin:str,            # Value for X-M2M-Origin header 
            device:Device,
           )->Device:

    if not device:
        raise Exception(f"update() requires parameter device as input.")

    #####
    # If the device info object is "dirty"
    if device.dirty_dvi:
        # Create the oneM2M dvi dictionary from the device
        dvi:dict = device.to_onem2m_dvi()

        # Re-create the dvi object of the device node.
        onem2m_dvis = device.onem2m.get('m2m:dvi', []) 
        #if device._init_with_dvi and onem2m_dvis:
        if onem2m_dvis:
            #to_delete_dvi_rn = f"{device.name}/{device.onem2m.get('m2m:dvi')[0].get('rn')}"
            to_delete_dvi_rn = f"{device.name}/{onem2m_dvis[0].get('rn')}"

            log.info(f"Deleting old managed dvi object {to_delete_dvi_rn} from DMO ...")
            dmo.delete_resource(api_client, origin, to_delete_dvi_rn)
        #else:
        #    log.warning(f"Onem2m node of device {device.name} does NOT have a device information object !!")    
            
        log.info(f"Adding new manged dvi object for device {device.name} in DMO ...")
        dmo.add_resource(api_client, origin=origin, r_type="managed-object", r_name=device.name, 
                            onem2m_body={"m2m:dvi": dvi})
        
    ####
    # Update credentials if they are "dirty"
    if device.dirty_credentials:
        deleted_flag = False

        # Delete existing credentials
        onem2m_crds = device.onem2m.get('m2m:crds', []) 

        # if device._init_with_crds and onem2m_crds:
        if onem2m_crds and onem2m_crds[0].get("ri", None): 
            to_delete_crds_rn = f"{device.name}/{onem2m_crds[0].get('rn')}"
            log.info(f"Deleting old device credentials {to_delete_crds_rn} from DMO ...")
            dmo.delete_resource(api_client, origin, to_delete_crds_rn)

            deleted_flag = True

        # Create the credentials 
        crds:dict = device.to_onem2m_crds()
        
        #log.info(f"Adding new credentials object for device {device.name} in DMO ...")
        #dmo.add_resource(api_client, origin=origin, r_type="managed-object", r_name=device.name, 
        #                 onem2m_body={"m2m:crds": crds})
        utils.retry_function(dmo.add_resource, 
                             args=[api_client], 
                             kwargs={
                                 "origin":origin, 
                                 "r_type":"managed-object", 
                                 "r_name":device.name, 
                                 "onem2m_body":{"m2m:crds": crds}
                             },
                             max_try_count=4, 
                             max_secs=15, 
                             sleep_before_first_try_secs=DELETE_CREDENTIALS_SLEEP_SECS if deleted_flag else -1,
                             sleep_after_error_secs=DELETE_CREDENTIALS_SLEEP_SECS,
                             error_return_value=log.error,
                             info_message="Adding new credentials object for device {device.name} in DMO ...",
                             info_logger=log.info, 
                             error_logger=log.error)
    ###
    # Update the labels in the node resource
    log.info(f"Updating labels of device {device.name} in DMO ...")
    dmo.update_resource_labels(api_client, origin, "nod", device.name, device.labels)

    # Reload the device and return it
    log.info(f"Reloading {device.name} from DMO ...")
    return get(api_client, origin, name=device.name)

def update(api_client:ApiClient,  # Client connection definition.
           origin:str=None,       # Value for X-M2M-Origin header 
           device:Device=None,
           name:str=None,
           id:str=None,
           **kwargs
          ):

    if device:
        return _update(api_client, origin, device=device)
    elif not name and not id:
        raise Exception(f"update() requires parameter device, name or id as input.")

    device = get(api_client, origin=origin, id=id, name=name)

    # Copy the keyword arguments into the device object
    device.set_from_kwargs(**kwargs)

    return _update(api_client, origin, device=device)

def create(api_client:ApiClient, # Client connection definition.
           origin:str=None,      # Value for X-M2M-Origin header 
           name:str=None,
           timeout_secs:int=10,
           **kwargs):

    # Verify if the name contains forbidden characters
    # MARK: WORK
    if not Device.validate_name(name):
        raise Exception(f"Invalid characters in name {name}. The name should look like {VALID_NAME_CHARS}")

    ## Build the credentials 
    credentials_id = kwargs.get("credentials_id", None)
    credentials_secret = kwargs.get("credentials_secret", None)

    if credentials_id and credentials_secret:
        node_creds = [f"psk-id:{credentials_id}",f"psk:{credentials_secret}"]
        node_creds_arg = {"nodeCredentials":node_creds}
    else:
        node_creds_arg = {}

    # Build the device information for the node
    '''
    "nodeDvi": {
        "rn": "device-info",
        "mgd": 1007,
        "man": "evel-corp",
        "dty": "world-destructor",
        "mod": "WD-15",
        "dlb": "some:label",
        "dc": "it destroys everything"
    },
    '''    
    dvi:dict = {
        "rn": DEVICE_INFO_RN,
        "mgd": DEVICE_INFO_MGD
    }
    
    for (arg_name, dvi_name, default) in (
                                          ("device_type",       "dty", ""), 
                                          ("location",          "loc",  None), #"geo:25.245470,51.454009"),
                                          ("label",             "dlb",  ""), 
                                          ("description",       "dc",   ""), 
                                          ("device_name",       "dvnm", ""), 
                                          ("firmware_version",  "fwv",  ""), 
                                          ("software_version",  "swv",  ""), 
                                          ("os_version",        "osv",  ""), 
                                          ("hardware_version",  "hwv",  ""), 
                                          ("protocols",         "ptl",  []),
                                          ("manufacturer",      "man",  ""), 
                                          ("model",             "mod",  ""), 
                                          ("sub_model",         "smod", ""), 
                                          ("country",           "cnty", "") 
                                         ):
        if (v := kwargs.get(arg_name, None)) == None:
            v = default
        if v != None:
            dvi[dvi_name] = v
        
    ###
    # Build labels 
    labels=[]
    if (props := kwargs.get("uplink_properties", None)):
        for (k, v) in props.items():
            labels.append(f"{UPLINK_PROPERTY_PREFIX}{k}:{v}")

    if (props := kwargs.get("device_properties", None)):
        for (k, v) in props.items():
            labels.append(f"{k}:{v}")

    if labels:
        labels_arg = {"labels":labels}
    else:
        labels_arg = {}

    # Step 1: Create device provisioning request which is execute asynchronously to 
    #         create the device in d2c. a
    log.info(f"Creating device provisioning request in DMO for {name} ...")
    d = device_provisioning.create_request(api_client,
                                           origin=origin,
                                           # Add the additional request data 
                                           nodeID=f"urn:gsma:imei:{name}",
                                           nodeResourceName=name,
                                           ICCID=kwargs.get("iccid", "89374121224379144333"),
                                           profile=DEVICE_PROFILES.SCS_LWM2M,
                                           **node_creds_arg, 
                                           ndvi=dvi,
                                           **labels_arg 
                                          )

    # Step 2: Wait for the device provisioning response until the job is executed
    #         succesfully or with errors. 
    '''
    POST of device provisioning request returns: 
        "dtiot:noPRt.resourceID": "67dc0c88c9a47b53f222656c"
        "dtiot:noPRt.resourceName": "device_provisioning_request_20250320-133933",

    GET of device provisioning reqeust contains: 
        "rn": "device_provisioning_request_20250320-133933",
        "ri": "67dc0c88c9a47b53f222656c"                        = dtiot:noPRt.resourceID

    GET of device provisioning respoonse contains:
        "rn": "responseToRequest-67dc0c88c9a47b53f222656c",      = "responseToRequest-"<request.ri>
        "reqRD": "67dc0c88c9a47b53f222656c",    
    '''
    requestId:str = utils.getValueFromDict(d, "dtiot:noPRt.resourceID", None, ".")
    responseName:str = f"responseToRequest-{requestId}" if requestId else None

    start_time=time.time()
        
    log.info(f"Waiting for response ...")
    found_response = False        
    while True:
        
        try:
            log.info(f"  trying to get response {responseName} from DMO ...")
            response = device_provisioning.get_responses(api_client, origin=origin, name=responseName)
            
            # If no response can be found 
            if not response:
                if time.time() > start_time + timeout_secs:
                    log.info(f"  ... not  found. Waiting for device provisioning response '{responseName}' timed out.")
                    break
                else:
                    log.info(f"  ... not found. Sleeping 1.0 seconcds ...")                    
                    time.sleep(1.0)
                    continue

            found_time = time.time()
            
            # Get the processing status from the response and test if it is ok.
            status = response.get("resSs", None)
            # "geEMe":"Total to be provisioned: 1; Successfully processed: 1; Errors: 0"
            general_message = response.get("geEMe", "No general processing message available")   
            individual_message = response.get("inEMs", "No individual processing message available")
            
            if status != 1:
                raise Exception(f"Device provisioninig request ended with status {status}.\n{general_message}\n{individual_message}")

            # We exit the loop to wait for the response object
            log.info(f"... found response after {found_time - start_time} seconds.")
            found_response = True
            break

        except Exception as ex:
            log.error(f"Error while getting the device provisioning response '{responseName}': {ex}")
            raise ex

    # Step 3: Even if we didn't find the provisioning request we try to load the device node from dmo.  
    if not found_response:
        log.warning("Could not find device provisioning response. Trying to get the new device anyway ...")
        
    device:Device = get(api_client, origin=origin, name=name)
    
    if not device:
        raise Exception(f"Could not get device {name} from DMO. Was the device provisioning request processed successfully?")    

    return device

def _inject_message(api_client:ApiClient, 
                    origin:str=None, 
                    msg_store:MessageStore=None,
                    id:str=None, 
                    name:str=None,
                    device:Device=None,
                    content=None,
                    content_type:str="text/plain",
                    contentEncoding:str="none",
                    **kwargs):    
    if content == None:
        return None
    elif msg_store == None:
        raise Exception(f"_inject_message() requires msg_store parameter.")
    elif not device and not name and not id:
        raise Exception(f"_inject_message() requires parameter device, name or id as input.")        
    
    # Determine the onem2m id of the encoding type
    contentEncodingId = utils.get_enum_value_by_name(dmo.CE, contentEncoding, default=dmo.CE.none.value)

    if content_type == "text/plain":
        content = str(content)
    elif content_type == "application/json":
        content = json.loads(content)
    else:
        raise Exception(f"Unsupported content type {content_type}.")
    
    body = {
        "m2m:cin": {
            "rn":    f"{uuid.uuid4()}",
            "con":   content,
            "cnf":   f"{content_type}:{contentEncodingId}"
        }
    }

    resource_name = f"device-communication/"
    if device:
        resource_name += device.name
    elif name:
        resource_name += name
    else:
        resource_name += id
    resource_name += f"/{msg_store.rn}"        
    
    response = dmo.add_resource(api_client, 
                                origin=origin, 
                                r_type="cin", 
                                r_name=resource_name,
                                onem2m_body=body)
    
    if not response:
        return None
    
    msg_id = response.get("m2m:cin", {}).get("ri", None)
    
    if not msg_id:
        log.error(f"Could not get id of message.")
        return None

    onem2m = dmo.get_resources(api_client, origin=origin, resourceId=msg_id)
    if not onem2m:
        log.error(f"Could not get message resource {msg_id}.")
        return None

    return Message(onem2m=onem2m)

def get_messages(api_client:ApiClient,
                 origin:str=None,   
                 name:str=None,    
                 id:str=None,
                 msg_store:MessageStore=None,
                 msg_name:str=None,
                 last:bool=False,
                 select:str=None,                 
                 format:str="CSV",             
                 limit:int=1000000000,    
                 offset:int=0     
                ):        
    if not msg_store:
        raise Exception("Function get_messages() requires msg_store as input parameter")
    elif not name and not id:
        raise Exception("Function get_messages() requires name or id as input parameter")

    log.info(f"Getting {msg_store.label} messages for device {name or id} from DMO ...")
    
    if id and not name:
        device = get(api_client, origin=origin, id=id)
        if not device:
            raise Exception(f"Could not load device with id {id}.")
        name = device.name

    # Build the resource name of the device message store
    resource_name = f"device-communication/{name}/{msg_store.rn}"

    # If a concrete message shall be get add its identfier to resource name
    if msg_name:
        resource_name += f"/{msg_name}"

    # If only the last message shall be get
    elif last:
        resource_name += "/la"
        limit=-1
    
    response = dmo.get_resources(api_client, 
                                origin=origin,
                                resourceName=resource_name,
                                limit=limit,
                                offset=offset)

    if not response:
        return None
    elif last:
        objects = Message(onem2m=response)
    elif response.get("con", None):
        objects = Message(onem2m=response)
    else:
        objects = [Message(onem2m=onem2m) for onem2m in response.get("m2m:cin", [])]

    if not select:
        return objects
    else:
        return dmo.select_from_response_body(objects, select=select, format=format)            

def delete_message(api_client:ApiClient,
                   origin:str=None,   
                   name:str=None,    
                   id:str=None,
                   msg_store:MessageStore=None,
                   msg_name:str=None,
                   msg_id:str=None
                  ):        
    # If the msg_id has been defined the message can be deleted without any other info
    if msg_id:
        log.info(f"Deleting message {msg_id} ...")
        dmo.delete_resource(api_client, origin=origin, id=msg_id)
        return        
    elif not msg_name:
        raise Exception("Function delete_message() requires parameter msg_name as input.")

    if not msg_store:
        raise Exception("Function delete_message() requires msg_store as input parameter")
    elif not name and not id:
        raise Exception("Function delete_message() requires name or id as input parameter")

    log.info(f"Deleting message {msg_name} from {msg_store.label} for device {name or id} from DMO ...")
    
    if id and not name:
        device = get(api_client, origin=origin, id=id)
        if not device:
            raise Exception(f"Could not load device with id {id}.")
        name = device.name

    # Build the resource name of the message to delete and delete it
    resource_name = f"device-communication/{name}/{msg_store.rn}/{msg_name}"
    dmo.delete_resource(api_client, origin=origin, name=resource_name)

    return None


import json
import logging
from argparse import ArgumentParser

import dtiot_d2c.d2c as d2c
import dtiot_d2c.dmo as dmo
from dtiot_d2c.cli.cli_command import ApiConnCmdLineInterface
from dtiot_d2c.d2c import CMDARGS, Device, DeviceGroup, OneM2MBase
from dtiot_d2c.d2c import utils as utils
from dtiot_d2c.d2c.get_response import GetResponse
from dtiot_d2c.dmo import dmoCLI as dmocli

log = logging.getLogger(__name__)
            
#####################################################################################                
class AddDeviceGroupCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Creates a new device group."
    
    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-n",required=True)
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-de")  # --description
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-ds")  #--devices
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-as")  #--applications

    def main(self, cmdargs=None):
            
        response = DeviceGroup.create(self.apiConn, 
                                      origin=cmdargs.origin,
                                      name=cmdargs.name,
                                      devices=cmdargs.devices,
                                      applications=cmdargs.applications,
                                      description=cmdargs.description
                                     )
        GetResponse(response=response).print()        
            
#####################################################################################
class GetDeviceGroupsCLI(ApiConnCmdLineInterface):
    
    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Returns the details of a device group."
    
    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-n")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-i")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-li")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-os")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-rci")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-s")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-f")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-sc")
                
    def main(self, cmdargs=None):

        format = cmdargs.format # if cmdargs.format else "CSV" if cmdargs.select else None

        response = DeviceGroup.get(self.apiConn, 
                                   origin=cmdargs.origin,
                                   id=cmdargs.id,
                                   name=cmdargs.name,
                                   limit=cmdargs.limit,
                                   offset=cmdargs.offset,
                                   resultContentIndocator=cmdargs.rci, 
                                   select=cmdargs.select,
                                   format=format)
        
        if not response:
            raise Exception(f"Device group {cmdargs.name or cmdargs.id} does not exist")

        GetResponse(response=response).print(cmdargs.sepchar)        
            
#####################################################################################
class UpdateDeviceGroupCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Updates a device group."

    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-o") 
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-n")  # --name
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-i")  # --id
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-de") # --description
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-ds") # --devices
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-as") # --applications
        
    def main(self, cmdargs=None):
            
        response = DeviceGroup.update(self.apiConn, 
                                      origin=cmdargs.origin,
                                      id=cmdargs.id,
                                      name=cmdargs.name,
                                      devices=cmdargs.devices,
                                      applications=cmdargs.applications,
                                      description=cmdargs.description
                                     )
        
        if not response:
            raise Exception(f"Device group {cmdargs.name or cmdargs.id} does not exist")
                
        GetResponse(response=response).print()    

#####################################################################################            
class DeleteDeviceGroupCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Deletes a device group."
    
    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-n")  # --name
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-i")  # --id
        
                
    def main(self, cmdargs=None):
        obj = DeviceGroup.delete(self.apiConn, 
                                 origin=cmdargs.origin, 
                                 id=cmdargs.id,
                                 name=cmdargs.name)

        if obj:
            print(str(obj))                                    

class ModifyDevicesCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Adds devices or removes devices to or from the device group."

    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-o") 
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-n", required=True)
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-addns")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-rmns")

    def main(self, cmdargs=None):
        add_names = utils.splitString(cmdargs.add_names, ",") if cmdargs.add_names else None
        remove_names = utils.splitString(cmdargs.remove_names, ",") if cmdargs.remove_names else None
            
        response = DeviceGroup.update_devices(self.apiConn, 
                                              origin=cmdargs.origin,
                                              name=cmdargs.name,
                                              add_names=add_names,
                                              remove_names=remove_names,
                                             )
            
        GetResponse(response=response).print()   
          
class ModifyApplicationsCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Returns the devices of the device group."

    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-o") 
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-n", required=True)
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-addns")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-rmns")

    def main(self, cmdargs=None):
        add_names = utils.splitString(cmdargs.add_names, ",") if cmdargs.add_names else None
        remove_names = utils.splitString(cmdargs.remove_names, ",") if cmdargs.remove_names else None
            
        response = DeviceGroup.update_applications(self.apiConn, 
                                                   origin=cmdargs.origin,
                                                   name=cmdargs.name,
                                                   add_names=add_names,
                                                   remove_names=remove_names,
                                                  )
        if not response:
            raise Exception(f"Device group {cmdargs.name or cmdargs.id} does not exist")
                    
        GetResponse(response=response).print()         
                    
        
class GetDeviceGroupDevicesCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Adds applications or removes applications to or from the device group."

    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-o") 
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-n", required=True)
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-s")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-f")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-sc")

    def main(self, cmdargs=None):
            
        device_group:DeviceGroup = DeviceGroup.get(self.apiConn, cmdargs.origin, cmdargs.name)

        #dics=[]
        #for device_id in device_group.deviceIds:
        #    device:Device = Device.get(self.apiConn, cmdargs.origin, )
        #GetResponse(response=response).print()                                   


import sys
import json
import time
import logging

from dtiot_d2c.cli.cli_command import ApiConnCmdLineInterface
import dtiot_d2c.dmo as dmo
from dtiot_d2c.dmo import dmoCLI as dmocli

log = logging.getLogger(__name__)
            
def print_flushed(s:str, end=None):
    print(s, end=end)
    sys.stdout.flush() 
    
#####################################################################################                
class PerformanceTest_GetResourceCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Runs a test to verify the performance of getting a resource."
    
    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-rn")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-ri")

        argsParser.add_argument("-n", "--number-of-gets", metavar="<integer>", 
                                dest="n", type=int, default=1,
                                help="Number of gets requests. Default is 1.")

                
    def main(self, cmdargs=None):
        
        if cmdargs.n <= 0:
            raise Exception(f"Parameters -n or --number-of-gets shall be greater than 0.")

        print(f"*** START ***")
        start_time = time.time()
        for i in range(cmdargs.n):
            print_flushed(f"{i + 1} ... ", end="")

            dmo.get_resources(self.apiConn, 
                              origin=cmdargs.origin, 
                              resourceId=cmdargs.ri, 
                              resourceName=cmdargs.rn)

        end_time = time.time()

        duration_total = end_time - start_time
        duration_per_get = duration_total / cmdargs.n 
        print("")
        print(f"*** END ***")
        print(f"  Resource:               {cmdargs.rn or cmdargs.ri}")
        print(f"  Number of get requests: {cmdargs.n}")
        print(f"  Total time:             {duration_total:.4f} seconds ")
        print(f"  Time per get request:   {duration_per_get:.4f} seconds")
        print("")




from io import DEFAULT_BUFFER_SIZE
import sys
import os
import os.path
import logging
import json
from pathlib import Path

import dtiot_d2c.cli.utils as utils
import dtiot_d2c.cli.CFG as CFG
from dtiot_d2c.cli.profile import Profiles

log = logging.getLogger(__name__)

class Configuration():

    def __init__(self, configFile=None, configDir=None, autoRead=True):
        self._file = None
        self._activeProfileFile = None
        self._profilesDirectory = None
        self._profiles = None

        if autoRead:
            self.read(configFile=configFile, configDir=configDir)
    
    def __str__(self):
        return json.dumps(self.getAsDict())
    
    def getAsDict(self):
        return {
            CFG.CEN_ACTIVE_PROFILE_FILE : self._activeProfileFile,
            CFG.CEN_PROFILE_DIR : self._profilesDirectory
        }

    def set(self, activeProfileFile=None, profilesDirectory=None):
        self._activeProfileFile = activeProfileFile
        self._profilesDirectory = profilesDirectory

    def getFile(self):
        return self._file

    def getActiveProfileFile(self):
        return os.path.basename(self._activeProfileFile)
        #return os.path.expandvars(s.replace("~", str(Path.home()))) if s else None
    def setActiveProfileFile(self, file):
        self._activeProfileFile = os.path.basename(file) if file else None
        #self._activeProfileFile = os.path.expandvars(file.replace("~", str(Path.home())))

    def getProfilesDirectory(self):
        s = self._profilesDirectory
        return os.path.expandvars(s.replace("~", str(Path.home()))) if s else None
    def setProfilesDirectory(self, directory):
        self._profilesDirectory = os.path.expandvars(directory.replace("~", str(Path.home())))

    def getProfiles(self):
        if not self._profiles:
            self._profiles = Profiles(config=self)

        return self._profiles
        
    def read(self, configFile=None, configDir=None):
        # get path of the configuration file
        cfgFile = Configuration.getPathToConfigFile(configFile=configFile, configDir=configDir)

        if not os.path.exists(cfgFile):
            raise Exception("Config file %s does not exist." % (cfgFile))

        log.debug("Reading config file %s" % (str(cfgFile)))

        # Read the configuration from the config file
        d = utils.readJsonFile(cfgFile)

        self.set(activeProfileFile=d[CFG.CEN_ACTIVE_PROFILE_FILE],
                 profilesDirectory=d[CFG.CEN_PROFILE_DIR])

        self._file = cfgFile

    def write(self, file=None, backup=False):
        if not file:
            file = self._file

        if backup:
            utils.backupFile(file)

        utils.writeJsonFile(self.getAsDict(), file)

        self._file = file

    ####################
    # Static methods
    ####################

    def getPathToConfigFile(configFile=None, configDir=None):
        '''
        if   config-file parameter is defined: use it
        elif config-dir  parameter is defined: use <config-dir>/config.json
        elif config-file envvar is defined:    use it
        elif config-dir  parameter is defined: use <config-dir>/config.json
        else                                   use ~/.ni/config.json
        '''
        if configFile:
            log.debug("Using config file from input parameters: %s" % (configFile))
            configFile = os.path.expandvars(configFile.replace("~", str(Path.home())))
        elif configDir:
            log.debug("Using config directory from input parameters: %s" % (configDir))
            configDir = os.path.expandvars(configDir.replace("~", str(Path.home())))
            configFile = "%s/%s" % (configDir, CFG.DEFAULT_CFGFILE_NAME)

        elif CFG.ENVVARNAME_CFGFILE in os.environ.keys():
            s = os.environ[CFG.ENVVARNAME_CFGFILE]
            log.debug("Using config file from environment: %s" % (s))
            configFile = os.path.expandvars(s.replace("~", str(Path.home())))

        elif CFG.ENVVARNAME_CFGDIR in os.environ.keys():
            s = os.environ[CFG.ENVVARNAME_CFGDIR]
            log.debug("Using config directory from environment: %s" % (s))
            configDir = os.path.expandvars(s.replace("~", str(Path.home())))
            configFile = "%s/%s" % (configDir, CFG.DEFAULT_CFGFILE_NAME)
        else:
            s = "~/%s/%s" % (CFG.DEFAULT_CFGDIR_NAME, CFG.DEFAULT_CFGFILE_NAME)
            log.debug("Using defaults for both config directory and config file name: %s" % (s))
            configFile = os.path.expandvars(s.replace("~", str(Path.home())))

        log.debug("configFile: %s" % (configFile))
        return configFile

    def create(configDir=None, overwrite=False):
        log.debug("=> Configuration.createConfiguration()")

        cfgFile = Configuration.getPathToConfigFile(configDir=configDir)

        log.debug("Creating confg file %s ..." % (cfgFile))

        if os.path.exists(cfgFile) and not overwrite:
            raise Exception("Configuration file %s already exists. User overwrite flag to overwrite the file." % (
                               cfgFile))
        
        # Create config directory if it does not exist
        cfgDirPath = os.path.dirname(cfgFile)
        if not os.path.exists(cfgDirPath):
            os.makedirs(cfgDirPath)


        config = Configuration(autoRead=False)
        config.setActiveProfileFile(None)
        config.setProfilesDirectory(os.path.join(cfgDirPath, CFG.DEFAULT_PROFILES_DIRNAME))
        config.write(cfgFile)

        # Create the profiles directory within the config directory and copy
        # the default files into it.

        profilesDirPath = "%s/%s" % (cfgDirPath, CFG.DEFAULT_PROFILES_DIRNAME)
        if not os.path.exists(profilesDirPath):
            os.makedirs(profilesDirPath)

        # fromDir = os.path.join(os.path.dirname(__file__), CFG.DEFAULT_PROFILES_DIRNAME)
        # for fromFile in utils.getFilesInDir(fromDir):
        #     fromFile = os.path.join(fromDir, fromFile)
        #     toFile = "%s/%s" % (profilesDirPath, os.path.basename(fromFile))
        #     log.debug("Copy file from %s to %s" % (fromFile,toFile))
        #     utils.copyFile(fromFile, toFile)

        return Configuration(configFile=cfgFile, autoRead=True)



#!/usr/bin/env python3

import argparse
import logging
import os
import os.path
import sys
import traceback

import dtiot_d2c.cli.CFG as CFG
import dtiot_d2c.cli.utils as utils
from dtiot_d2c.cli.cli_command import ApiConnCmdLineInterface
from dtiot_d2c.cli.cli_command import CLICommands as clicmds
from dtiot_d2c.cli.cli_command import (ConfigCmdLineInterface,
                                       ProfileCmdLineInterface)
from dtiot_d2c.cli.config import Configuration
from dtiot_d2c.cli.profile import Profile
from dtiot_d2c.cli.utils import color

log = logging.getLogger(__name__)

WAIT_FOR_DBG_CLIENT=True

def printUsage():
    print("Usage: d2c MAJOR-COMMAND [MINOR-COMMAND] [OPTIONS] [PARAMETERS]")

def printSupportedCommands():
    print("Supported commands are:")
    print("  --help")
    for n in clicmds.getCommandNames():
        print("  %s" % (n))

# Build commands help
class SmartFormatter(argparse.HelpFormatter):
    def _split_lines(self, text, width):
        if text.startswith('R|'):
            return text[2:].splitlines()  
        return argparse.HelpFormatter._split_lines(self, text, width)

def _main():

    '''
    ni COMMAND [options] [parameters]
    Options are:
    -h, --help: 
    -l, --loglevel:
    --debug
    ''' 
    argsNum = len(sys.argv)
    arg1 = sys.argv[1] if argsNum > 1 else None

    # Create the command line parse with the basic command line options ...
    argsParser = argparse.ArgumentParser(prog="d2c", 
                                         description="D2C command line interface", 
                                         formatter_class=SmartFormatter)

    # Build the command's help
    cmdsHelp = "R|"
    for grpname in clicmds.getCommandGroupNames():
        cmdgrp = clicmds.getCommandGroup(grpname)

        cmdsHelp = "%s%s%s%s%s\n" % (cmdsHelp, color.UNDERLINE, color.BOLD, cmdgrp["name"], color.END)

        for cmd in cmdgrp["commands"]:
            cmdNames = cmd.getName()
            for a in cmd.getAliasNames():
                cmdNames = "%s, %s" % (cmdNames, a)

            cmdsHelp = "%s%s: %s\n" % (cmdsHelp, cmdNames, cmd.callGet1LineHelp())

        cmdsHelp = "%s\n" % (cmdsHelp)

    argsParser.add_argument("command", nargs='?', help=cmdsHelp)

    argsParser.add_argument("command-options", nargs="*", 
                            help="Command options.")

    # If no command line arguments are passed, display help messages and exit
    if len(sys.argv) == 1:
        argsParser.print_help()
        return 1
    
    # Remove all command line args behind the command name to  let parse the basic args
    # If we don't do this the argument parser creates an error because we can not 
    # declare the command specific arguments here. We don't know them.

    # Copy original command line args
    argvBackup = []
    for arg in sys.argv:
        argvBackup.append(arg)    

    # Search for the command name in the arguments ...
    cmdNames = []

    for cmd in clicmds.getCommands():
        cmdNames.append(cmd.getName())
        #print("A>%s<" % (str(cmd.getName())))
        for a in cmd.getAliasNames():
            #print("B>%s<" % (str(a)))
            cmdNames.append(a)

    cmdIdx = None
    for idx in range(1, len(sys.argv)):
        if sys.argv[idx] in cmdNames:
            cmdIdx = idx
            break

    # Remove all command line args behind the command name
    if cmdIdx:
        while len(sys.argv) > (cmdIdx + 1):
            sys.argv.pop(-1)

    #  Parse the command lines arguments
    cmdargs = argsParser.parse_args()

    ###
    # Initialize the command and execute it
    cmd = clicmds.getCommand(cmdargs.command)

    if not cmd:
        print("Error: unknown command %s.\n" % (cmdargs.command), file=sys.stderr)
        argsParser.print_help()
        return 1

    log.debug("Selected command %s of type %s." % (cmdargs.command, str(type(cmd))))

    # Re-create the command line parser for command from the stored command line args
    sys.argv = argvBackup[cmdIdx:]
    
    argsParser = argparse.ArgumentParser(prog="d2c %s" % (cmdargs.command), 
                                         description=cmd.get1LineHelp())

    argsParser.add_argument("-l", 
                            choices = ["INFO", "info", "WARN", "warn", "DEBUG", "debug"],
                            dest="log_level", default="WARN",
                            help="Log level. Possible values  are INFO, info, WARN, warn, DEBUG, debug.") 

    argsParser.add_argument("--debug", dest="debug", action="store_const", const=True, default=False,
                            help="If set, than debug output is printed. Same like -l debug")

    argsParser.add_argument("--info", dest="info", action="store_const", const=True, default=False,
                            help="If set, than info output is printed. Same like -l info")

    argsParser.add_argument("--outfile", metavar="<filepath>", dest="outfile", default="-",
                            help="File into which the log outputs shall be written additionaly.")

    # add configuration arguments if required by command
    if isinstance(cmd, ConfigCmdLineInterface):
        argsParser.add_argument("--config-directory", metavar="<directory>", dest="configDir", 
                                help="Directory which contains the configuration file %s." % (CFG.DEFAULT_CFGFILE_NAME))

        argsParser.add_argument("--config-file", metavar="<file>", dest="configFile", 
                                help="Path to the configuration file. If set than --config-directory is ignored.")

    # add profile argument if required by command
    if isinstance(cmd, ProfileCmdLineInterface):
        argsParser.add_argument("-p", "--profile-file", metavar="<file>", dest="profileFile",
                                help="Path to the profile file. If not set the configured active profile is ued.")

    cmd.addCmdLineArgs(argsParser)

    cmdargs = argsParser.parse_args()

    if cmdargs.debug:
        cmdargs.log_level = "DEBUG"
        logging.basicConfig(force=True, level=cmdargs.log_level.upper(), format="%(asctime)s | %(levelname)s | %(name)s | %(message)s")                                
    else:
        if cmdargs.info:
            cmdargs.log_level = "INFO"

        logging.basicConfig(force=True, level=cmdargs.log_level.upper(), format="%(asctime)s | %(levelname)s | %(message)s")                                

   # Debug the parsed command line options
    for e in dir(cmdargs):
        if not e.startswith("_"):
            log.debug("%-20s = %s" % (e, str(getattr(cmdargs, e))))

    # Add an addtional logging output file if requested by command line option
    if cmdargs.outfile and len(cmdargs.outfile) > 0 and cmdargs.outfile != "-":
        file_handler = logging.FileHandler(cmdargs.outfile)
        file_handler.setLevel(level=cmdargs.log_level.upper())
        formatter = logging.Formatter("%(asctime)s | %(levelname)s | %(name)s | %(message)s")                                
        file_handler.setFormatter(formatter)
        logging.getLogger().addHandler(file_handler)

    # If the current command requires a configuration check if the 
    # configuration file is available.
    if isinstance(cmd, ConfigCmdLineInterface):
        cfgFile = Configuration.getPathToConfigFile(configFile=cmdargs.configFile, 
                                                    configDir=cmdargs.configDir)
        if not os.path.exists(cfgFile):
            raise Exception("Config file %s does not exist." % (cfgFile))

        cmd.config = Configuration(configFile=cfgFile)

    # If the current command requires a profile load the profile
    if isinstance(cmd, ProfileCmdLineInterface):

        if not cmd.config:
            raise Exception("Configuration hasn't been loaded.")

        if not cmdargs.profileFile:
            profileFile = os.path.join(cmd._config.getProfilesDirectory(), cmd._config.getActiveProfileFile())
        else:
            profileFile = cmdargs.profileFile

        cmd.profile = Profile(file=profileFile)

    # If the current command requires a Nokia Impact API connection

    if isinstance(cmd, ApiConnCmdLineInterface):
        if not cmd.profile:
            raise Exception("No profile has been loaded.")

        cmd.apiConn = CFG.createApiConnection(cmd.profile)
    
    ###
    # Execut the command ...
    cmd.callMain(cmdargs)

    return 0

def main():
    exitCode = 0

    try:
        
        exitCode = _main()
        
    except Exception as ex:
        if log.root.level == logging.DEBUG:
            traceback.print_exc()
        else:
            utils.print_stderr(f"Error: {ex}")
            utils.print_stderr("Use \"--debug\" command line option to see more info.")
            #traceback.print_exc()
        exitCode = 1
    
    finally:
        sys.exit(exitCode)
            
if __name__ == '__main__':
    main()    



import json
import logging
from datetime import datetime
from typing import List

from pydantic import BaseModel, Field, PrivateAttr

import dtiot_d2c.d2c.consts as d2c_consts
import dtiot_d2c.dmo as dmo
from dtiot_d2c.d2c import utils as utils
from dtiot_d2c.d2c.onem2m_base import OneM2MBase
from dtiot_d2c.dmo import ApiClient

from .consts import *

log = logging.getLogger(__name__)

class Application(OneM2MBase):
    _application_type: str = PrivateAttr(default=None)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        # Initialize the application type
        self._application_type = self._get_label(APPL_TYPE_LABEL_NAME, default_value=APPL_TYPE_DEFAULT)

    @property
    def urls(self)->List[str]:
        return self.onem2m.get("poa", []) 

    @urls.setter
    def urls(self, value:List[str]):
        urls_1 = self.urls
        self.onem2m["poa"] = utils.integrate_lists(urls_1, value)

    @property
    def type(self):
        return self._get_label(f"{APPL_TYPE_LABEL_NAME}")

    @property
    def description(self):
        return self._get_label(f"{DESCRIPTION_LABEL_NAME}")

    @description.setter
    def description(self, value): 
        labels = {f"+{DESCRIPTION_LABEL_NAME}":value}
        self.integrate_labels(labels)
        
    ### 
    # connection properties

    def set_connection_properties(self, props:dict):
        prefix = CONN_PROPS_LABEL_PREFIXS.get(self._application_type, APPL_TYPE_DEFAULT)
        self._set_prefixed_labels(prefix, props)

    def delete_connection_properties(self):
        prefix = CONN_PROPS_LABEL_PREFIXS.get(self._application_type, APPL_TYPE_DEFAULT)
        self._delete_prefixed_labels(prefix)

    def set_connection_property(self, key, value):
        prefix = CONN_PROPS_LABEL_PREFIXS.get(self._application_type, APPL_TYPE_DEFAULT)
        self._set_prefixed_label(prefix, key, value)
        
    def del_connection_property(self, key):
        prefix = CONN_PROPS_LABEL_PREFIXS.get(self._application_type, APPL_TYPE_DEFAULT)
        self._del_perfixed_property(prefix, key)

    @property    
    def connectionProperties(self)->dict:   
        prefix = CONN_PROPS_LABEL_PREFIXS.get(self._application_type, APPL_TYPE_DEFAULT)
        return self._get_prefixed_labels(prefix)
    
    @connectionProperties.setter
    def connectionProperties(self, props:dict):
        self.set_connection_properties(props)
            
    ### regular functions
    #
    def model_dump(self,  *args, **kwargs):
        md = super().model_dump(*args, **kwargs)
        for key in ["type",
                    "urls", 
                    "connectionProperties",
                    "description",
                   ]:
            md[key] = getattr(self, key)
        del md["labels"]
        return md
        
    def set_from_kwargs(self, **kwargs):
        super().set_from_kwargs(**kwargs)
        
        #### Copy simple string argumentsw
        for (paramName, attrName) in (("description", "description"),
                                     ):
            if not (value := kwargs.get(paramName, None)):
                continue
            setattr(self, attrName, value)
                    
        # Copy list args into the object
        for key in ["urls",
                   ]:
            if not (value := kwargs.get(key, None)):
                continue

            if type(value) == list:
                setattr(self, key, value)
            elif type(value) == str:
                setattr(self, key, utils.splitString(value, ","))
            else:
                log.error(f"Unsupported type {type(value)} for argument {key}={value}")

        ### 
        # Copy the connection properties into the labels of the application
        self.integrate_labels(kwargs.get("connection_properties", None), 
                              self.del_connection_property,
                              self.set_connection_property,
                              self.set_connection_properties)
                  
    def to_onem2m(self, **kwargs)->dict:
        d = {}
        for (attr_name, onem2m_path) in [["urls", "poa"],
                                        ]:
            d[onem2m_path] = getattr(self, attr_name)

        d["lbl"] = self.onem2m.get("lbl", [])

        # Add the d2c version
        d2c_version_lbl = f"{d2c_consts.D2C_SYSTEM_VERSION_LABEL}:{d2c_consts.D2C_SYSTEM_VERSION}"
        if d2c_version_lbl not in d["lbl"]:
            d["lbl"].append(d2c_version_lbl)

        # Add the application type
        appl_type = kwargs.get("application_type", APPL_TYPE_WEBHOOK)
        appl_type_lbl = f"{APPL_TYPE_LABEL_NAME}:{appl_type}"
        if appl_type_lbl not in d["lbl"]:
            d["lbl"].append(appl_type_lbl)
            
        # add the additional keyword arguments
        for (key, value) in kwargs.items():
            d[key] = value

        return {"m2m:ae":d}

    ######################################################################
    # Static public functions
    ######################################################################
    def get_create_onem2m_dict(name:str, **kwargs)->dict:
        application_type = kwargs.get("application_type", APPL_TYPE_DEFAULT)

        labels = kwargs.get("labels", {})
        labels[APPL_TYPE_LABEL_NAME] = application_type
        labels[d2c_consts.D2C_SYSTEM_VERSION_LABEL] = d2c_consts.D2C_SYSTEM_VERSION
        labels[DESCRIPTION_LABEL_NAME] = kwargs.get("description", None)
        
        # Add the connection properties to the labels
        props = kwargs.get("connection_properties", None) or {}
        for k, v in props.items():
            n = CONN_PROPS_LABEL_PREFIXS.get(application_type)
            labels[f"{n}{k}"] = v

        return  {
            "m2m:ae": {
                "api": f"N{name}",
                "rn": f"{name}",
                "lbl": dmo.convert_dict_to_onem2m_labels(labels),
                "poa": kwargs.get("urls", []),
            }
        }
        
    def get(api_client:ApiClient,  
            origin:str=None,   
            id:str=None,    
            name:str=None,
            select:str=None,                 # Comma separated list of element names which shall be selected.
            format:str="CSV",                # Output format. Possible values are CSV and JSON
            resultContentIndocator:str=None, # Result content indicator. Supported types are RCIs (see above).
            limit:int=1000000000,            # Max number of objects to return.
            offset:int=0
           ) -> dict:  

        return get(api_client, origin, id, name, resultContentIndocator=resultContentIndocator,
                   limit=limit, offset=offset, select=select, format=format)
        
    def create(api_client:ApiClient,
               origin:str=None,     
               name:str=None,
               **kwargs):    
        return create(api_client, origin=origin, name=name, **kwargs)
    
    def delete(api_client:ApiClient, 
               origin:str=None, 
               name:str=None, 
               id:str=None
              ):
        return delete(api_client, origin=origin, name=name, id=id)
    
    def update(api_client:ApiClient, 
               origin:str=None, 
               device_group=None,  # DeviceGroup
               id:str=None,
               name:str=None,
               **kwargs
              ):  
        return update(api_client, origin, 
                      device_group=device_group, 
                      id=id,
                      name=name,
                      **kwargs)  
        
################################################################
################################################################
################################################################
#MARK: IMPLEMENTATIONS

def _get(api_client:ApiClient,  
         origin:str=None,  
         id:str=None,    
         name:str=None,
         resultContentIndocator:str=None, # Result content indicator. Supported types are RCIs (see above).
         limit:int=1000000000,            # Max number of objects to return.
         offset:int=0,
         select:str=None,                 # Comma separated list of element names which shall be selected.
         format:str="CSV"                 # Output format. Possible values are CSV and JSON
        ) -> dict:        
    
    if not name and not id:
        response = dmo.get_resources(api_client, 
                                     origin=origin,
                                     resourceType="ae", 
                                     limit=limit,
                                     offset=offset,
                                     resultContentIndocator=resultContentIndocator)
    else:
        response = dmo.get_resources(api_client, 
                                     origin=origin,
                                     #resourceType="ae", 
                                     resourceId=id,
                                     resourceName=name,
                                     limit=limit,
                                     offset=offset,
                                     resultContentIndocator=resultContentIndocator)      
    if not response:
        return None
        
    if not name and not id:
        object = [Application(onem2m=onem2m) for onem2m in response]
    else:
        object = Application(onem2m=response)  
                
    # Return the plain objects
    if not select:
        return object
    else:
        return dmo.select_from_response_body(object, select=select, format=format)

def get(api_client:ApiClient,
        origin:str=None,  
        id:str=None, 
        name:str=None,    
        select:str=None,                 
        format:str="CSV",             
        limit:int=1000000000,  
        offset:int=0,      
        resultContentIndocator:str=None, 
       ):        
    if name or id:
        log.info(f"Getting application {name or id} from DMO ...")
    else:
        log.info(f"Getting applications from DMO ...")        

    return _get(api_client, origin=origin, id=id, name=name, select=select, format=format, 
                resultContentIndocator=resultContentIndocator, limit=limit, offset=offset)

def create(api_client:ApiClient, 
           origin:str=None,      
           name:str=None,
           **kwargs):

    # Only create the application if id doesn't exist
    
    log.info(f"Testing if the application {name} already exists ...")
    if (app := get(api_client, origin, name=name)):
        log.warning(f"  Application {name} already exists.")
        return app

    log.info(f"Creating application {name} in DMO ...")
    onem2m = Application.get_create_onem2m_dict(name, **kwargs)
    dmo.add_resource(api_client, origin=origin if origin else name, r_type="ae", onem2m_body=onem2m)

    return get(api_client, origin, name=name)

def delete(api_client:ApiClient, 
           origin:str=None, 
           name:str=None, 
           id:str=None
          ):
    if not name and not id:
        raise Exception(f"delete() requires parameter name or id as input.")
    
    log.info(f"Deleting application {name or id} from DMO ...")
    return dmo.delete_resource(api_client, origin=origin, name=name, id=id)

def _update(api_client:ApiClient, 
            origin:str,           
            object:Application,
           )->OneM2MBase:

    if not object:
        raise Exception(f"_update() requires parameter object as input.")

    # Create the oneM2M from the object
    onem2m:dict = object.to_onem2m()

    log.info(f"Updating application {object.name} in DMO ...")        
    dmo.update_resource(api_client, origin, name=object.name, onem2m_body=onem2m)

    # Reload the device group and return it
    return get(api_client, origin=origin, name=object.name)

def update(api_client:ApiClient,  
           origin:str=None,       
           object:OneM2MBase=None,
           id:str=None,
           name:str=None,
           **kwargs
          ):
    
    if object:
        return _update(api_client, origin, object)
    elif not name and not id:
        raise Exception(f"update() requires parameter object or name or id as input.")

    object = get(api_client, origin, id, name)  

    # Copy the keyword arguments into the object
    object.set_from_kwargs(**kwargs)
            
    return _update(api_client, origin, object)





import json
import logging
from datetime import datetime
from typing import List

from pydantic import BaseModel, Field

import dtiot_d2c.d2c.consts as d2c_consts
import dtiot_d2c.dmo as dmo
from dtiot_d2c.d2c import utils as utils

log = logging.getLogger(__name__)

class OneM2MBase(BaseModel):
    onem2m:dict = Field(default_factory=dict)

    def _get_child_array_attribute(self, child_name, attr_name, default_value=None):
        child = self.onem2m.get(child_name, [])
        return child[0].get(attr_name, default_value) if child else default_value

    def _set_child_array_attribute(self, child_name, attr_name, value):
        child = self.onem2m.get(child_name, None)
        if not child:
            child = self.onem2m[child_name] = []
            child.append({})
        child[0][attr_name] = value
    
    @property
    def id(self):
        return self.onem2m.get("ri", None) if self.onem2m else None

    @property
    def name(self):
        return self.onem2m.get("rn", None) if self.onem2m else None

    @property
    def labels(self)->dict:
        return dmo.convert_onem2m_labels_to_dict(self.onem2m.get("lbl", [])) if self.onem2m else {}

    def _set_labels(self, value:dict):
        self.onem2m["lbl"] = dmo.convert_dict_to_onem2m_labels(value)

    @labels.setter
    def labels(self, value:dict):
        self._set_labels(value)

    def _get_label(self, label_name, default_value=None):
        return self.labels.get(label_name, default_value)
    
    def _set_label(self, label_name, label_value):
        d = self.labels
        if label_value != None:
            d[label_name] = label_value
        elif label_name in d.keys():
            del d[label_name]
        self.labels = d
        
    def _del_label(self, label_name):
        self._set_label(label_name, None)        

    def integrate_labels(self, labels, del_label_func=None, set_label_func=None, set_labels_func=None):
        if not del_label_func:
            del_label_func = self._del_label
        if not set_label_func:
            set_label_func = self._set_label
        if not set_labels_func:
            set_labels_func = self._set_labels

        operation = "replace"

        if labels:        
            for key in labels.keys():
                if key[0] in (d2c_consts.UPDT_LABS_OPERATION_INDICATOR_ADD, d2c_consts.UPDT_LABS_OPERATION_INDICATOR_DEL):
                    operation = "integrate"
                    break

        if operation == "integrate":
            for key, value in labels.items():
                if key[0] == d2c_consts.UPDT_LABS_OPERATION_INDICATOR_DEL or value == None:
                    del_label_func(key[1:])
                elif key[0] == d2c_consts.UPDT_LABS_OPERATION_INDICATOR_ADD:
                    set_label_func(key[1:], value)
                else:
                    set_label_func(key, value)
        else:
            if type(labels) == str:
                labels = json.loads(labels)

            if type(labels) == dict:
                set_labels_func(labels)

            elif labels != None:
                raise Exception(f"Type {type(labels)} for onem2m labels not supported.")        
            
    @property
    def creationTime(self)->datetime:
        return dmo.convert_onem2m_timestamp_to_datetime(self.onem2m.get("ct", "20250101T000000,000000"))

    @property
    def lastModificationTime(self)->datetime:
        return dmo.convert_onem2m_timestamp_to_datetime(self.onem2m.get("lt", "20250101T000000,000000"))

    def model_dump(self,  *args, **kwargs):
        md = super().model_dump(*args, **kwargs)
        del md["onem2m"]
        for key in ["id", "name", "labels", "creationTime", 
                    "lastModificationTime"
                   ]:
            md[key] = getattr(self, key)
        return md
        
    def set_from_kwargs(self, **kwargs):
        self.integrate_labels(kwargs.get("labels", None))

    def to_onem2m(self)->dict:
        raise Exception("Function to_onem2m() not implemented.")

    def __str__(self):
        return str(self.model_dump())

    def to_json(self)->str:
        """Returns the object as a json string.
        Returns:
            str: JSON string
        """
        return json.dumps(self.dict(), sort_keys=False, indent=4, cls=utils.DatetimeJsonEncoder)
    
    ### 
    # Manage prefixed labels
    def _get_prefixed_labels(self, prefix:str)->dict:
        d = {}
        for (k, v) in self.labels.items():
            if k.startswith(prefix):
                d[k[len(prefix):]] = v
        return d

    def _get_unprefixed_labels(self, ignore_prefixes:List[str], system_labels:List[str],
                               prefix_separator_char="_"
                               )->dict:
        d = {}
        for (k, v) in self.labels.items():
            if k in system_labels:
                continue

            idx = k.find(prefix_separator_char)
            if idx >= 0 and k[0:idx] in ignore_prefixes:
                continue
    
            d[k] = v

        return d

    def _delete_prefixed_labels(self, prefix:str):
        old_labels = self.labels
        new_labels = {}
        for (k, v) in old_labels.items():
            if not k.startswith(prefix):
                new_labels[k] = old_labels[k]
        self.labels = new_labels

    def _set_prefixed_labels(self, prefix:str, props:dict):
        if props == None:
            return
        
        self._delete_prefixed_labels(prefix)
        
        for (key, value) in props.items():
            self._set_label(f"{prefix}{key}", value)   

    def _set_prefixed_label(self, prefix:str, key:str, value):
        self._set_label(f"{prefix}{key}", value)

    def _del_perfixed_property(self, prefix:str, key:str):
        self._del_label(f"{prefix}{key}")

    
    
    



import json
import logging
from enum import Enum
from typing import Union

from pydantic import BaseModel

import dtiot_d2c.d2c.utils as utils
from dtiot_d2c.d2c.onem2m_base import OneM2MBase

log = logging.getLogger(__name__)

class GetResponseType(Enum):
    ANY = 0
    ONEM2M_BASE_OBJECT = 1
    LIST_OF_ONEM2M_BASE_OBJECTS = 2
    DICT = 3
    LIST_OF_DICTS = 4
    STR = 5
    LIST_OF_STRS = 6
    LIST_OF_LISTS = 7
    NONE = 99
    UNSUPPORTED_TYPE=100

class GetResponse(BaseModel):
    response:Union[None, dict, list, str, OneM2MBase] 
    type: GetResponseType = GetResponseType.ANY
    
    def __init__(self, **data):
        super().__init__(**data)    
        self._init_type_from_response(self.response)
        
    def _init_type_from_response(self, response):
        if response is None:
            self.type = GetResponseType.NONE
        elif isinstance(response, OneM2MBase): 
            self.type = GetResponseType.ONEM2M_BASE_OBJECT
        elif isinstance(response, dict):
            self.type = GetResponseType.DICT
        elif isinstance(response, str):
            self.type = GetResponseType.STR
        elif isinstance(response, list):
            if response and isinstance(response[0], OneM2MBase): 
                self.type = GetResponseType.LIST_OF_ONEM2M_BASE_OBJECTS
            elif response and isinstance(response[0], dict):
                self.type = GetResponseType.LIST_OF_DICTS
            elif response and isinstance(response[0], str):
                self.type = GetResponseType.LIST_OF_STRS
            elif response and isinstance(response[0], list):
                self.type = GetResponseType.LIST_OF_LISTS
        else:
            self.type = GetResponseType.UNSUPPORTED_TYPE
                
    def print(self, sepchar:str=";"):
        #print(f"type: {type(self.response)}")
        #print(f"type: {self.type}")

        if not self.response or self.type == GetResponseType.NONE:
            return 
        elif self.type == GetResponseType.UNSUPPORTED_TYPE:
            print(f"Type {self.type} not supported. Cannot print result {self.result}.")
            return

        if self.type == GetResponseType.ONEM2M_BASE_OBJECT:
            print(self.response.to_json())
        elif self.type == GetResponseType.DICT:
            print(json.dumps(self.response, indent=4, cls=utils.DatetimeJsonEncoder))
        elif self.type == GetResponseType.STR:
            print(self.response)
        elif self.type == GetResponseType.LIST_OF_ONEM2M_BASE_OBJECTS:
            cnt=0
            print("[")
            for obj in self.response:
                print(obj.to_json(), end="")
                cnt += 1
                if cnt < len(self.response):
                    print(",")
            print("")
            print("]")
        elif self.type == GetResponseType.LIST_OF_DICTS:
            cnt=0
            print("[")
            for obj in self.response:
                print(json.dumps(obj, indent=4, cls=utils.DatetimeJsonEncoder), end="")
                cnt += 1
                if cnt < len(self.response):
                  print(",")
            print("")
            print("]")
        elif self.type == GetResponseType.LIST_OF_STRS:
            print(sepchar.join(self.response))
        elif self.type == GetResponseType.LIST_OF_LISTS:
            for obj in self.response:
                line = sepchar.join(obj)
                print(line)








import glob
import importlib
import json
import logging
import os
import os.path
from tkinter import *
from tkinter import messagebox

import dtiot_d2c.cli.CFG as CFG
from dtiot_d2c.cli.profile import Profile

log = logging.getLogger(__name__)


from .ProfileEditor import ProfileEditor
from .ProfileSelector import ProfileSelector


class ProfileManagerGui():
    def __init__(self,  config):
        self._config = config
        self._profiles = config.getProfiles()
        self._selectedProfile = None
        self._activatedProfile = None
        self._newProfile = None

        rootWnd = Tk()
        rootWnd.title("NICLI Profile Manager")
        self._rootWnd = rootWnd

        # Build the left side of the window with the profile selector
        leftFrame = Frame(rootWnd)
        
        
        self._prfSelector = ProfileSelector(parent=leftFrame,
                                            profiles=self._profiles,
                                            controller=self,
                                            selectionChangedCallback=self.profileHasBeenSelectedCallBack)

        frame = self._buildCommandsFrame(leftFrame)
        self._prfSelector.pack(side=TOP, expand=TRUE, fill=BOTH, padx=0, pady=0)
        frame.pack(side=BOTTOM, expand=False, fill=X, padx=0, pady=0)  
    
        # Create right side of window with the profile editor
        self._prfEditor = ProfileEditor(parent=rootWnd, profileManager=self)

        # Pack the left and the right side of the window
        leftFrame.pack(side=LEFT, expand=False, fill=BOTH, padx=10, pady=10)
        self._prfEditor.pack(side=LEFT, expand=True, fill=BOTH, padx=10, pady=10)

    def getProfileCfg(self)->dict:
        return CFG.PROFILE_CONFIG
        
    def _buildCommandsFrame(self, parent):
        frame = Frame(parent)

        button = Button(frame, text="ACTIVATE", command=self.activateProfile, state="disabled", width=12)
        button.pack(side=RIGHT, expand=True, padx = 10, pady = 10)
        self._activateButton = button

        button = Button(frame, text="NEW", command=self.newProfile, state="normal", width=12)
        button.pack(side=RIGHT, expand=True, padx = 10, pady = 10)
        self._newButton = button

        button = Button(frame, text="COPY", command=self.copyProfile, state="disabled", width=12)
        button.pack(side=RIGHT, expand=True, padx = 10, pady = 10)
        self._copyButton = button

        button = Button(frame, text="DELETE", command=self.deleteProfile, state="disabled", width=12)
        button.pack(side=RIGHT, expand=True, padx = 10, pady = 10)
        self._deleteButton = button

        return frame

    def activateProfile(self, profile=None):
        if not profile:
            profile = self._selectedProfile

        file = profile.getFile()

        self._config.setActiveProfileFile(file)
        self._config.write()
        
        self._prfSelector.showActivatedProfile(profile)
        self._activateProfile = profile

    def copyProfile(self):
        if self._selectedProfile:
            self.newProfile(copyProfile=self._selectedProfile)
        else:
            self.newProfile()

    def newProfile(self, copyProfile=None):
        profile = Profile(profileCfg=self.getProfileCfg())
        self._prfEditor.setProfile(profile)
        self._newProfile = profile

    def deleteProfile(self):
        if not self._selectedProfile:
            return
        
        name = self._selectedProfile.getBaseFilename()

        rv = messagebox.askokcancel(
                title="Delete Profile", 
                message="Do you really want to delete profile %s?" % (name),
                icon=messagebox.QUESTION)
        
        if not rv:
            return

        file = self._selectedProfile.getFile()
        if not os.path.exists(file):
            raise Exception("Cannot delete profile. File %s does not exist." % (file))
        else:
            os.remove(file)

        self._selectedProfile = None
        self._profiles.loadProfiles()
        self._prfEditor.setProfile(None)
        self._prfSelector.updateProfilesList()

    def _updateButtonsState(self):
        if self._selectedProfile:
            state = NORMAL
        else:
            state = DISABLED

        for b in [self._activateButton, self._copyButton, self._deleteButton]:
            b.configure(state=state)

    def profileHasBeenSelectedCallBack(self, profile):
        log.info("=> profileHasBeenSelectedCallBack(profile=%s)" % (str(profile)))
        self._prfEditor.setProfile(profile)
        self._selectedProfile = profile

        self._updateButtonsState()

    def saveProfile(self, profile):
        log.info("=> saveProfile() ...")
        log.info(json.dumps(json.loads(str(profile)),indent=4))

        file = profile.getFile()

        if file and len(file.strip()) > 0:
            profile.write()

        else:
            # Build the filename of the profile
            fname = ""
            for e in self.getProfileCfg()["filenameElements"]:
               fname += profile._profileDict.get(e, e)
            dir = self._config.getProfilesDirectory()

            if os.path.exists(os.path.join(dir, fname)):
                raise Exception("A profile with the same account and project already exist. Choose a different ones!")

            profile.write(directory=dir, file=fname)
            self._profiles.addProfile(profile)
            self._prfEditor.setProfile(profile)
            self._prfSelector.updateProfilesList(select=profile)

    def run(self):
        self._rootWnd.mainloop()



from fastapi import Request
from . import api_utils

from  .datamodel import HealthzResponse

import logging
log = logging.getLogger(__name__)

def init():
    '''
    Initializes the module
    '''
    pass

def handleRequest(request:Request):
    api_utils.logRequest(request)
    return HealthzResponse(message = "OK")




import importlib
import imp
import logging
import os
import os.path
import json
from pathlib import Path
from datetime import datetime

log = logging.getLogger(__name__)

class color:
   PURPLE = '\033[95m'
   CYAN = '\033[96m'
   DARKCYAN = '\033[36m'
   BLUE = '\033[94m'
   GREEN = '\033[92m'
   YELLOW = '\033[93m'
   RED = '\033[91m'
   BOLD = '\033[1m'
   UNDERLINE = '\033[4m'
   END = '\033[0m'

def importModule(moduleName, moduleFile):
    """
    Imports a module from source code and returns the exected module
    @param moduleName Name of the module
    @param moduleFile File of the source code
    @return Excecuted module is returned 
    @throws Exception in case of an error
    """
    #log.debug("=> importModule(moduleName=%s, moduleFile=%s)" % (moduleName, moduleFile))

    # Python 2.7
    module = imp.load_source(moduleName, moduleFile)

    # Python 3.5
    #spec = importlib.util.spec_from_file_location(moduleName, moduleFile)
    #module = importlib.util.module_from_spec(spec)
    #spec = spec.loader.exec_module(module)

    return module

def newInstanceFromClass(moduleName, className):
    '''
    Creates an instance of a class from a module and class name string.
    @param moduleName Name of the module
    @param className Name of the class in the module.
    @return Instance of class
    @throws Exception in case of an error.
    '''
    log.debug("=> createClassInstance(moduleName=%s, className=%s)" % (moduleName, className))

    mod = importlib.import_module(moduleName)
    cl = getattr(mod, className, None)

    if cl == None:
        raise Exception("Could not get class object %s from module %s." % (className, moduleName))

    obj = cl()

    if obj == None :
        raise Exception("Could not create instance of class %s from module %s." % (className, moduleName))
    
    log.debug("<= createClassInstance()");
    return obj

def splitString(str, sepChar):
    a = []
    for s in str.split(sepChar):
        a.append(s.strip())
    return a

def isDict(obj):
    if not obj:
        return False
    else:
        return True if obj.__class__ == {}.__class__ else False

def isArray(obj):
    if not obj:
        return False
    else:
        return True if obj.__class__ == [].__class__ else False

# Return the local IP
import socket
def getLocalIp():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    
    IP=None
    
    try:
        s.connect(('10.254.254.254', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        s.close()
    
    return IP    

# Class to executues a function every x seconds 
import threading
import time
import traceback

class FunctionExecutor(threading.Thread):
    def __init__(self, cronSecs, function, *args, **kwargs):
        super().__init__()
        self._cronSecs = cronSecs
        self._function = function
        self._args = args
        self._kwargs = kwargs

    def run(self):
        while True:
            try:
                time.sleep(self._cronSecs)
                self._function(*self._args, **self._kwargs)
            except Exception as ex:
                log.error(str(ex))
                traceback.print_exc()
                
#MARK: execCmd
from enum import Enum
class ReturnType(Enum):
    TEXT = 1
    JSON_AS_DICT = 2
    LINES_AS_LIST = 3

def execCmd(cmd, returnType : ReturnType = ReturnType.TEXT):
    log.debug("Exec %s" % (cmd))
        
    cmdIn = os.popen(cmd)

    line = cmdIn.readline()
    s = ""
    while line:
        if line.startswith("b\'"):
            line = line.strip()
            line = line[2:-1] + "\n"
        s = s + line
        line = cmdIn.readline()    

    log.debug("\n%s" % (s.strip()))                                  

    if returnType == ReturnType.JSON_AS_DICT:
        return json.loads(s)
    elif returnType == ReturnType.LINES_AS_LIST:
        return splitString(s, "\n")
    else:
        return s                
    
########################################################################
#MARK: date function
########################################################################
from datetime import datetime

def fnow(format:str="%Y-%m-%d %H:%M:%S")->str:
    return datetime.now().strftime(format)    


import json
import logging
from datetime import datetime
from typing import List

import dtiot_d2c.dmo as dmo
from dtiot_d2c.d2c import utils as utils
from dtiot_d2c.d2c.onem2m_base import OneM2MBase
from dtiot_d2c.dmo import ApiClient
from pydantic import BaseModel, Field

from .consts import *

log = logging.getLogger(__name__)

'''
 {
            "rn": "23fe29a2-0302-4791-ba79-ac7330048a5f",
            "con": "Hallo von Roland",
            "cnf": "text/plain:0",
            "ty": 4,
            "cs": 16,
            "st": 0,
            "cr": "Cd2ccli4",
            "ri": "67ebf327c8a381a4e97e2668",
            "pi": "67ebb9d5c8a381a4e97e2614",
            "ct": "20250401T140735,303000",
            "lt": "20250401T140735,303000"
}
'''
class Message(OneM2MBase):
    
    @property
    def content(self):
        return self.onem2m.get("con", None) 

    @content.setter
    def content(self, value):
        self.onem2m["con"] = value
        
    @property
    def contentInfo(self):
        return self.onem2m.get("cnf", "text/plain:0") 

    @contentInfo.setter
    def contentInfo(self, value):
        self.onem2m["cnf"] = value

    @property
    def contentType(self):
        if not (ci := self.contentInfo):
            return None
        elif (idx := ci.find(":")) < 0:
            return 
        else:
            return ci[:idx]

    @property
    def contentEncoding(self):
        if not (ci := self.contentInfo):
            return None
        elif (idx := ci.find(":")) < 0:
            return 
        else:
            try:
                return utils.get_enum_name_by_value(dmo.CE, int(ci[idx+1:]), default="")                
            except Exception as ex:
                log.error(f"Could not convert content encoding {ci[idx+1:]}: {ex}")
                return ""

    @property
    def contentSize(self):
        return self.onem2m.get("cs", "text/plain:0") 

    @contentSize.setter
    def contentSize(self, value):
        self.onem2m["cs"] = value

    def model_dump(self,  *args, **kwargs):
        md = super().model_dump(*args, **kwargs)
        for key in ["content", 
                    "contentInfo",
                    "contentSize",
                    "contentType",
                    "contentEncoding"
                   ]:
            md[key] = getattr(self, key)
        del md["labels"]
        return md
        
  
        


import json
import logging
import threading
import time

import requests

from . import html_utils as hutils
from . import utils

log = logging.getLogger(__name__)

class _CFG:
    dmoBasepath:str = "device-management-orchestrator" 
    minApiTokenValiditySecs:int = 10  # Min validity secs of an api token. If the left validity is 
                                      # is smaller the token is refreshed.

def cfg(**kwargs):
    global _CFG
    for (k, v) in kwargs.items():
        if hasattr(_CFG, k):
            setattr(_CFG, k, v)
        else:
            log.warning(f"Ignoring unknown configuration element {k}.")

class ApiToken:
    '''
    API token after successfull client authentication.
    '''
    def __init__(self, info:dict):
        '''
        info: 
            {
                "access_token": "eyJhbGciOiJSUzI1NiIsInR5cCIgOiAiSldUIiwia2lkIiA6ICJaS1U2QjVQNXQ2aTcwRzh0R2YwZEdsejBBVlN5cXhZYms5THRQN3N2cjBvIn0.eyJleHAiOjE3MDUzMTE5MjYsImlhdCI6MTcwNTMxMTYyNiwianRpIjoiOWRjODEyNmEtMGI2ZC00OTRkLWEwYzItY2YxMmY2ZTRhYTZlIiwiaXNzIjoiaHR0cHM6Ly9pYW0tYnItaW90Lm15aW90LWQuY29tL2F1dGgvcmVhbG1zL2lvdGh1YiIsInN1YiI6IjI2MjNiMmRjLTc2MmEtNGE4Mi1hMDZiLWEwZTgzNTNlMWRjOSIsInR5cCI6IkJlYXJlciIsImF6cCI6IjQ2M2VhOGEzLTQ0OGEtNDBhZC05MzRiLWYxMzYxNTgzOGQ2ZSIsImFjciI6IjEiLCJyZWFsbV9hY2Nlc3MiOnsicm9sZXMiOlsicm8lY29ubmVjdGlvbi1tYW5hZ2VtZW50LW9yY2hlc3RyYXRvciIsInJ3JWNvbm5lY3Rpb24tbWFuYWdlbWVudC1vcmNoZXN0cmF0b3IiLCJybyVkZXZpY2UtbWFuYWdlbWVudC1vcmNoZXN0cmF0b3IiLCJydyVkZXZpY2UtbWFuYWdlbWVudC1vcmNoZXN0cmF0b3IiXX0sInNjb3BlIjoicHJvZmlsZSBlbWFpbCIsInRlbmFudHMiOlsiZW9zIl0sImVtYWlsX3ZlcmlmaWVkIjpmYWxzZSwiY2xpZW50SG9zdCI6Ijc5LjIxMC4xODYuMTc5IiwiY2xpZW50SWQiOiI0NjNlYThhMy00NDhhLTQwYWQtOTM0Yi1mMTM2MTU4MzhkNmUiLCJwcmVmZXJyZWRfdXNlcm5hbWUiOiJzZXJ2aWNlLWFjY291bnQtNDYzZWE4YTMtNDQ4YS00MGFkLTkzNGItZjEzNjE1ODM4ZDZlIiwiY2xpZW50QWRkcmVzcyI6Ijc5LjIxMC4xODYuMTc5In0.GthxWKymbpRVN0UVVuYGHGykMMleIx1E512hpBLPyCM82sY1QJvTzqx-Zqd6XTlVlFTzueAAT5ybRVl63uQlNqQisdAqwkbwF4qMwEawCpoN3UnXlzVBnp0s9uF81UCDQOcGXbDilaySYviuxJicFTKMbpsKsCbAPJNKte2FJ-QqRTFgAx6_wDA5pVlm4aRUsm6KwYU6I9nl7rAfAEaFpQCw1tttxuh5jVqkE-uoUQbwfrhE8NegztILD6TAL3bmIs_c4sUP-1XkxiIdOy5Y7QrDb8E_ozo-inMoskkAsq5d62Du_1l1LIhxtbgPkF6ctSA8q0shyxR19Y_QCfUeOw",
                "expires_in": 300,
                "refresh_expires_in": 0,
                "token_type": "Bearer",
                "not-before-policy": 0,
                "scope": "profile email"
            }
        '''
        self._info = info
        self._tokenTime = time.time()

    @property
    def access_token(self):
        return self._info.get("access_token", None)

    @property
    def validity_secs(self):
        return self._info.get("expires_in", 0) - (time.time() - self._tokenTime)


class ApiClient:
    '''
    API client credentials.
    '''
    
    def __init__(self, api_url:str, origin:str, get_token_url:str, client_id:str, client_secret:str):
        self._api_url:str = api_url
        self._origin:str = origin 
        self._get_token_url:str = get_token_url
        self._client_id:str = client_id
        self._client_secret:str = client_secret
        self._apiToken:ApiToken = None
        self._refreshTokenLock = threading.Lock()

    @property
    def key(self):
        return f"{self._tenant_name}:{self._client_id}:{self._client_secret}"

    @property
    def origin(self):
        return self._origin
    
    @property
    def api_url(self):
        return self._api_url           

    @property
    def api_url_unstructured(self):
        '''
        This is a dirty hack to make out of the regular, structured, api url which contains as the last
        path element the CseId (tenant) an unstructured api url to access resources by its it and not
        by its name
        '''
        l = self.api_url.split("/")
        return "/".join(l[:-1]) + "/~/" + l[-1]
    
    @property
    def get_token_url(self):
        return self._get_token_url

    @property
    def client_id(self):
        return self._client_id
    
    @property
    def client_secret(self):
        return self._client_secret   

    @property
    def token(self) -> ApiToken:
        return self._apiToken

    @property
    def bearer_token(self) -> str:
        
        if not self._apiToken:
            self._apiToken = self.auth()

        # Refresh the token
        elif self._apiToken.validity_secs < _CFG.minApiTokenValiditySecs:
            self._refreshTokenLock.acquire()
            try:
                log.info(f"Refresh access token for {self._client_id} ...")
                self._apiToken = self.auth()
            except Exception as ex:
                log.error(f"{ex}")
            finally:
                self._refreshTokenLock.release()            

        return self._apiToken.access_token
    
    def auth(self) -> ApiToken:
        '''
        Performs API authentication and returns the ApiToke.
                
        curl --request POST "https://playground.spacegate.telekom.de/auth/realms/default/protocol/openid-connect/token" \
        --header 'Content-Type: application/x-www-form-urlencoded' \
        --header 'Accept: application/x-www-form-urlencoded, application/json' \
        --header 'uthorization: Basic t-iot-hub--aag7zrqole--056e86c6-8b22-4c7c-a4fa-be9451987feb 0ca42ee6-9c07-4081-a536-9c3ceec2ed80' \
        --header 'Authorization: Basic dC1pb3QtaHViLS1hYWc3enJxb2xlLS0wNTZlODZjNi04YjIyLTRjN2MtYTRmYS1iZTk0NTE5ODdmZWI6MGNhNDJlZTYtOWMwNy00MDgxLWE1MzYtOWMzY2VlYzJlZDgw' \
        --data 'grant_type=client_credentials' \

        '''
        url = self.get_token_url
        hutils.log_url(log.debug, "POST", url)

        headers = {
            "Content-Type" :"application/x-www-form-urlencoded",
            "Accept": "application/x-www-form-urlencoded, application/json",
            "Authorization": f"Basic {utils.to_basic_auth(self._client_id, self._client_secret)}"
        }
        hutils.log_headers(log.debug, headers)

        body = "grant_type=client_credentials"
        hutils.log_body(log.debug, body)
       
        response = requests.post(url=url, headers=headers, data=body)
        hutils.log_response(log.debug, response)          
        
        '''
        {
            "access_token":"eyJhbGciOiJSUzI1NiIsInR5cCIgOiAiSldUIiwia2lkIiA6ICItQ2oyakVyejN0b1d0Y1FZVWtzRFB3aGFXZHQyS3AteWpZMTduRjBXdjZvIn0.eyJleHAiOjE3NDIzODE2NzUsImlhdCI6MTc0MjM4MTM3NSwianRpIjoiZTM0MWJhZmMtM2Q1Yi00MmYyLWJmMmUtYzlkZDgwY2Y5ZjMyIiwiaXNzIjoiaHR0cHM6Ly9wbGF5Z3JvdW5kLnNwYWNlZ2F0ZS50ZWxla29tLmRlL2F1dGgvcmVhbG1zL2RlZmF1bHQiLCJzdWIiOiJmOGIyODgxZi00MTQ1LTQ3ZmYtYTBiZC05M2ExYTBkOWMxZmQiLCJ0eXAiOiJCZWFyZXIiLCJhenAiOiJ0LWlvdC1odWItLWFhZzd6cnFvbGUtLTA1NmU4NmM2LThiMjItNGM3Yy1hNGZhLWJlOTQ1MTk4N2ZlYiIsImFjciI6IjEiLCJzY29wZSI6ImNsaWVudC1vcmlnaW4gcHJvZmlsZSBlbWFpbCIsImNsaWVudEhvc3QiOiIxMDAuMTAyLjE0NC44NSIsImNsaWVudElkIjoidC1pb3QtaHViLS1hYWc3enJxb2xlLS0wNTZlODZjNi04YjIyLTRjN2MtYTRmYS1iZTk0NTE5ODdmZWIiLCJlbWFpbF92ZXJpZmllZCI6ZmFsc2UsIm9yaWdpblpvbmUiOiJzcGFjZSIsInByZWZlcnJlZF91c2VybmFtZSI6InNlcnZpY2UtYWNjb3VudC10LWlvdC1odWItLWFhZzd6cnFvbGUtLTA1NmU4NmM2LThiMjItNGM3Yy1hNGZhLWJlOTQ1MTk4N2ZlYiIsImNsaWVudEFkZHJlc3MiOiIxMDAuMTAyLjE0NC44NSIsIm9yaWdpblN0YXJnYXRlIjoiaHR0cHM6Ly9wbGF5Z3JvdW5kLnNwYWNlZ2F0ZS50ZWxla29tLmRlIn0.AZByWXAthRPPy0O-c07_gHdvzHILZYEw1AFJlejBKL2ZxeL2YIZyDaj15FNk29cQWT8gK5CizsODKUqA2MPV47s50PtbH0AOKnFe203zgxrsOxAwUiRa_y0TKga9cIVjFN8llqDyJUCmCAMOIKdLtvul1H_k8sHx3BTjBpbQBhaNkMOugn1yHVwNQVgyXkiThBypPImJImiDu2zGqui8WLZGGi2tAHIeDhjZq7g1ane0aUUkzWCcXh_fOk9ecch0B9GBmk5RoDnbL_Lo51cZWk2zRjr2P7TdQQzG4ZZ8kO-fsw_Wz6rePFKQcGhFF44FikIM-Knf8xJFENNWIPCRqQ",
            "expires_in":300,
            "refresh_expires_in":0,"token_type":
            "Bearer","not-before-policy":0,
            "scope":"client-origin profile email"
        }                
        '''       
       
        if not response.ok:
            response.raise_for_status()

        d = json.loads(response.content)

        self._apiToken = ApiToken(d)        

        return self._apiToken

class _ApiClients:
    '''
    Container to managbe APIClient instances.
    '''
    def __init__(self):
        self._clients = {}
        
    def addClient(self, client:ApiClient):
        '''
        Add a client to the container.
        '''
        self._clients[client.key] = client
        return client
        
    def getClient(self, key:str) -> ApiClient:
        '''
        Returns a client instance by its key.
        
        :key:str: The key of the client as returned by the clients'
                  key property.
        '''
        return self._clients.get(key, None)

# Create the singelton instance of the ApiClients container
ApiClients = _ApiClients()
    
    



import os
import os.path
import logging
import json
import dtiot_d2c.cli.utils as utils
import dtiot_d2c.cli.CFG as CFG

log = logging.getLogger(__name__)

#-------------------------------------------------------------------------------
class Profile:
    def __init__(self, 
                 file=None, 
                 account=None, project=None, 
                 apiHost="api.scs.iot.telekom.com", apiPort=443, 
                 apiUsername=None, apiPassword=None, apiTenant=None, 
                 applicationUrl=None,
                 copyFromProfile=None,
                 profileCfg=None):

        self._file = file
        self._profileDict:dict = {}

        if self._file:
            self.read()
        elif copyFromProfile:
            self.copyFrom(copyFromProfile)
        elif profileCfg:
            for c in profileCfg["elements"]:
                self._profileDict[c["name"]] = c["default"]

    def copyFrom(self, other):
        self._profileDict = other._profileDict

    def __str__(self):
        return json.dumps(self._profileDict)

    def getAsDict(self):
        return self._profileDict

    def set(self, profileDict):
        self._profileDict = profileDict

    def getFile(self):
        return self._file
    def setFile(self, file):
        self._file = file

    def getElement(self, name, default=None):
        return self._profileDict.get(name, default)
    def setElement(self, name, value, type="string"):
        if type == "int":
            self._profileDict[name] = int(value)
        elif type == "float":
            self._profileDict[name] = float(value)
        else:            
            self._profileDict[name] = value

    def getBaseFilename(self):
        if self._file:
            return os.path.basename(self._file)
        else:
            return "%s_%s.%s" % (self._account, self._project, CFG.PROFILE_FILE_EXT)

    def buildFilename(self, profileCfg:dict)->str:
        '''
        Builds the filename according to the passed profile configuration.
        '''
        fname = ""
        for e in profileCfg["filenameElements"]:
            fname += self._profileDict.get(e, e)
        return fname
    
    def read(self, file=None):
        if not file and self._file:
            file = self._file

        if not os.path.exists(file):
            raise Exception("Profile file %s does not exist." % (file))

        log.debug("Reading profile file %s" % (file))

        # Read the profile from file
        d = utils.readJsonFile(file)

        for (k, v) in d.items():
            self._profileDict[k] = v

        self._file = file

    def write(self, directory=None, file=None):
        if not file:
            file = self._file
        else:
            if directory:
                file = os.path.join(directory, file)
            else:
                raise Exception("Cannot write profile %s. Directory required." % (file))

        utils.writeJsonFile(self._profileDict, file)

        self._file = file
        
#-------------------------------------------------------------------------------
class Profiles:
    def __init__(self, directory=None, config=None, autoLoad=True):

        if config and not directory:
            directory = config.getProfilesDirectory()

        self._directory = directory
        self._profiles = []

        if autoLoad:
            self.loadProfiles(directory)

    def loadProfiles(self, directory=None):
        if not directory:
            directory = self._directory

        if not directory:
            log.warn("Profiles directory is None")
            return None
        elif not os.path.exists(directory):
            raise Exception("Profiles directory %s does not exist." % (directory))            

        profiles = {}
        for file in utils.getFilesInDir(directory):
            if file.endswith(CFG.PROFILE_FILE_EXT):
                try:
                    profile = Profile(file=os.path.join(directory, file))
                    profiles[file] = profile
                except Exception as ex:
                    log.error("Could not load profile %s: %s" % (file, str(ex)))

        self._directory = directory
        self._profiles = profiles

    def getProfileKeys(self):
        return self._profiles.keys()

    def getProfiles(self, sorted=False):
        a = []
        for k in self._profiles.keys():
            a.append(self._profiles[k])

        if sorted:
            def getProfiles_getBaseFilename(profile):
                return profile.getBaseFilename()
            a.sort(key=getProfiles_getBaseFilename)

        return a

    def addProfile(self, profile):
        file = profile.getFile()

        if not file:
            raise Exception("Profile cannot be added. It has no file.")
        
        self._profiles[file] = profile

    def getProfile(self, file):
        return self._profiles.get(file)




import json
import logging
import sys
import time
from typing import List

import dtiot_d2c.d2c.test.utils as tutils
import dtiot_d2c.dmo as dmo
from dtiot_d2c.cli.cli_command import ApiConnCmdLineInterface
from dtiot_d2c.d2c import CMDARGS
from dtiot_d2c.d2c import utils as utils
from dtiot_d2c.d2c.device import Device
from dtiot_d2c.d2c.device.consts import *
from dtiot_d2c.d2c.get_response import GetResponse
from dtiot_d2c.d2c.test.testcli_base import TestCase, TestCLI_Base, TestSet
from dtiot_d2c.dmo import ApiClient
from dtiot_d2c.dmo import dmoCLI as dmocli

log = logging.getLogger(__name__)
log_level = log.getEffectiveLevel() 

def verify_device_communication(api:ApiClient, name:str):
    '''
    device-communication/<name> = 
    {
        "mbs": 1000000,
        "mni": 100,
        "rn": "device01-112",
        "ty": 3,
        "st": 0,
        "cbs": 0,
        "cni": 0,
        "cr": "CDevice-Provisioning",
        "ri": "67ffeee60aced48199fa7087",
        "pi": "67f641e2bbad596bd91446e3",
        "ct": "20250416T175446,849000",
        "lt": "20250416T175446,849000",
        "m2m:cnt": [
            {
                "mbs": 1000000,
                "mni": 100,
                "rn": "outgoing-msg",
                "ty": 3,
                "st": 0,
                "cbs": 0,
                "cni": 0,
                "cr": "CDevice-Provisioning",
                "ri": "67ffeee60aced48199fa7089",
                "pi": "67ffeee60aced48199fa7087",
                "ct": "20250416T175446,902000",
                "lt": "20250416T175446,902000",
                "m2m:sub": [
                    {
                        "rn": "protocol-adapter-subscription",
                        "nct": 1,
                        "enc": {
                            "net": [
                                1,
                                2,
                                3,
                                4
                            ]
                        },
                        "nu": [
                            "http://d2c-protocol-adapter-scs-lwm2m.d2c.svc.cluster.local/node/TMNL.D2CDEV.DEV_D2C_ROOT_002.d2c-dev-3/device01-112/downlink-message"
                        ],
                        "su": "http://d2c-protocol-adapter-scs-lwm2m.d2c.svc.cluster.local:80/node/TMNL.D2CDEV.DEV_D2C_ROOT_002.d2c-dev-3/device01-112/downlink-message",
                        "ty": 23,
                        "ri": "67ffeee70aced48199fa7091",
                        "pi": "67ffeee60aced48199fa7089",
                        "ct": "20250416T175447,341000",
                        "lt": "20250416T175447,341000"
                    }
                ]
            },
            {
                "mbs": 1000000,
                "mni": 100,
                "rn": "received-msg",
                "ty": 3,
                "st": 0,
                "cbs": 0,
                "cni": 0,
                "cr": "CDevice-Provisioning",
                "ri": "67ffeee70aced48199fa708d",
                "pi": "67ffeee60aced48199fa7087",
                "ct": "20250416T175447,015000",
                "lt": "20250416T175447,015000"
            },
            {
                "mbs": 1000000,
                "mni": 100,
                "rn": "sent-msg",
                "ty": 3,
                "st": 0,
                "cbs": 0,
                "cni": 0,
                "cr": "CDevice-Provisioning",
                "ri": "67ffeee60aced48199fa708b",
                "pi": "67ffeee60aced48199fa7087",
                "ct": "20250416T175446,958000",
                "lt": "20250416T175446,958000"
            }
        ]
    }        
    '''
    ###
    # Check the toplevel device-communication container of the device
    n = f"device-communication/{name}"
    def operation():
        onem2m_dict = dmo.get_resources(api_client=api, resourceName=n)
        if not onem2m_dict:
            raise Exception(f"Couldn't get resource {n}.")
        else:
            return onem2m_dict
    onem2m_dict:dict = tutils.exec_test(f"Getting device communication container {n}", operation)        

    ###
    # Check the device-communication/<name>/outgoing-msg container (downlink inbound)
    n = f"device-communication/{name}/outgoing-msg"
    def operation():
        onem2m_dict = dmo.get_resources(api_client=api, resourceName=n)
        if not onem2m_dict:
            raise Exception(f"Couldn't get resource {n}.")
        else:
            return onem2m_dict
    onem2m_dict:dict = tutils.exec_test(f"Getting device communication container {n}", operation)        

    ###
    # Is there an outgoing subscription to forward messages to the southbound adapter ?
    n = f"device-communication/{name}/outgoing-msg/protocol-adapter-subscription"
    def operation():
        onem2m_dict = dmo.get_resources(api_client=api, resourceType="subscription", resourceName=n)
        if not onem2m_dict:
            raise Exception(f"Couldn't get subscription {n}.")
        else:
            return onem2m_dict
    onem2m_dict:dict = tutils.exec_test(f"Getting protocol adapter subscription {n}", operation)      
    
    ###        
    # Check the device-communication/<name>/received-msg container (uplink inbound)
    n = f"device-communication/{name}/received-msg"
    def operation():
        onem2m_dict = dmo.get_resources(api_client=api, resourceName=n)
        if not onem2m_dict:
            raise Exception(f"Couldn't get resource {n}.")
        else:
            return onem2m_dict
    onem2m_dict:dict = tutils.exec_test(f"Getting device communication container {n}", operation)               

    ###        
    # Check the device-communication/<name>/sent-msg container (downlink outbound)
    n = f"device-communication/{name}/sent-msg"
    def operation():
        onem2m_dict = dmo.get_resources(api_client=api, resourceName=n)
        if not onem2m_dict:
            raise Exception(f"Couldn't get resource {n}.")
        else:
            return onem2m_dict
    onem2m_dict:dict = tutils.exec_test(f"Getting device communication container {n}", operation)               
    
def verify_device(api: ApiClient, name:str, device_props:dict):
    '''
    {
        "ni": "urn:gsma:imei:device01-103",
        "lbl": [
            "profile:SCS-lwM2M",
            "ICCID:89374121317959919924",
            "UplinkProperty_address:Kuckhoffstr 114A, 13156 Berlin",
            "lastMaintenance:25.3.2025 13 00"
        ],
        "rn": "device01-103",
        "ty": 14,
        "ri": "67ffe2e3eab987515f7c3484",
        "pi": "d2c-dev-3",
        "ct": "20250416T170331,008000",
        "lt": "20250416T170335,841000",
        "m2m:crds": [
            {
                "dc": "cred",
                "mgd": 1029,
                "rn": "credentials",
                "ty": 13,
                "ri": "67ffe2e3eab987515f7c348a",
                "pi": "67ffe2e3eab987515f7c3484",
                "ct": "20250416T170331,436000",
                "lt": "20250416T170331,436000",
                "crid": "device01-103",
                "crse": "Q3hhc21FTUw3NFNiUlIzUw=="
            }
        ],
        "m2m:sub": [
            {
                "rn": "device-update-subscription",
                "nct": 1,
                "enc": {
                    "net": [
                        1,
                        2,
                        3,
                        4
                    ]
                },
                "nu": [
                    "http://d2c-device-provisioning.d2c.svc.cluster.local/provisioning/node"
                ],
                "su": "http://d2c-device-provisioning.d2c.svc.cluster.local:80/provisioning/node",
                "ty": 23,
                "ri": "67ffe2e3eab987515f7c3488",
                "pi": "67ffe2e3eab987515f7c3484",
                "ct": "20250416T170331,325000",
                "lt": "20250416T170331,325000"
            }
        ],
        "m2m:dvi": [
            {
                "rn": "device01-103",
                "mgd": 1007,
                "loc": "geo:25.245470,51.454009",
                "dty": "Pretty Device",
                "dlb": "this is a label",
                "dvnm": "device01-103",
                "fwv": "fw 1.0.1",
                "swv": "sw 2.0.1",
                "osv": "os 3.0.1",
                "hwv": "2024.1 PC r1",
                "ptl": [
                    "LWM2M"
                ],
                "man": "PSsystec",
                "mod": "Smartbox Mini",
                "smod": "NB-IoT",
                "cnty": "DE",
                "ty": 13,
                "ri": "67ffe2e7eab987515f7c3490",
                "pi": "67ffe2e3eab987515f7c3484",
                "ct": "20250416T170335,053000",
                "lt": "20250416T170335,053000"
            }
        ]
    }        
    '''
    def operation():
        onem2m_dict = dmo.get_resources(api_client=api, resourceName=name)
        if not onem2m_dict:
            raise Exception(f"Node resource {name} couldn't be get after creation.")
        else:
            return onem2m_dict
    onem2m_dict:dict = tutils.exec_test(f"Getting ae {name}", operation)

    ###
    # Check node labels
    for (key, val) in [["profile", DEVICE_PROFILES.SCS_LWM2M],
                        ["ICCID", device_props.get("iccid", None)],
                        *[(k, v) for (k, v) in device_props.get("device_properties", {}).items()],
                        *[(f"UplinkProperty_{k}", v) for (k, v) in device_props.get("uplink_properties", {}).items()]
    ]:
        if val == None:
            continue
        elem_path = "lbl"
        elem_value = f"{key}:{val}"
        def operation():
            if not tutils.call_has_element_value(onem2m_dict, elem_path, elem_value):
                raise Exception(f"Value of element {elem_path} does not meet required value {elem_value}.")
        tutils.exec_test(f"Verifing {elem_path} for {elem_value}", operation)
                
    ###
    # Check dvi elements
    for (dvi_elem, cfg_elem) in [["man" ,"manufacturer"], 
                                    ["mod" ,"model"],
                                    ["smod" ,"sub_model"],
                                    ["hwv" ,"hardware_version"],
                                    ["cnty" ,"country"],
                                    #["dc" ,"description"],
                                    ["dlb" ,"label"],
                                    ["dty" ,"device_type"],
                                    ["dvnm" ,"device_name"],
                                    ["fwv" ,"firmware_version"],
                                    ["swv" ,"software_version"],
                                    ["osv" ,"os_version"],
                                    ["loc" ,"location"],
                                    #["ptl" ,"protocols"]
    ]:
        elem_path = f"m2m:dvi[0].{dvi_elem}"
        elem_value = device_props.get(cfg_elem, "__UNDEFINED__")

        if elem_value == "__UNDEFINED__":
            continue

        def operation():
            if not tutils.call_has_element_value(onem2m_dict, elem_path, elem_value):
                raise Exception(f"Value of element {elem_path} does not meet required value {elem_value}.")
        tutils.exec_test(f"Verifing {elem_path} for {elem_value}", operation)

    # elem_path = f"m2m:dvi[0].dvnm"
    # elem_value = name
    # def operation():
    #     if not tutils.call_has_element_value(onem2m_dict, elem_path, elem_value):
    #         raise Exception(f"Value of element {elem_path} does not meet required value {elem_value}.")
    # tutils.exec_test(f"Verifing {elem_path} for {elem_value}", operation)

    # Check credentials
    for (crds_elem, cfg_elem) in [["crid" ,"credentials_id"], 
                                  ["crse" ,"credentials_secret"],
    ]:
        elem_path = f"m2m:crds[0].{crds_elem}"
        elem_value = device_props.get(cfg_elem, "__UNDEFINED__")

        if elem_value == "__UNDEFINED__":
            continue

        def operation():
            if not tutils.call_has_element_value(onem2m_dict, elem_path, elem_value):
                raise Exception(f"Value of element {elem_path} does not meet required value {elem_value}.")
        tutils.exec_test(f"Verifing {elem_path} for {elem_value}", operation)

    return None

def delete_device(api:ApiClient, name:str):
    def operation():  
        if (application := Device.get(api, name=name)):
            Device.delete(api, name=name)
    tutils.exec_test(f"Deleting device {name}", operation)
    return None
class TestCase_CreateDevice(TestCase):
    def get_name(self):
        return "Create Device"
        
    def run(self):
        name = self.cfg.get("name")

        if not (create_props := self.cfg.get("create_props", None)):
            raise Exception(f"Cannot find element create_props in configuration.")

        def operation():    
            response = Device.create(self.api, name=name, **create_props)

            if not response:
                raise Exception(f"Device {name} hasn't been created")
        
            #if log_level <= tutils.INFO:
            #    GetResponse(response=response).print()      

        tutils.exec_test(f"Creating device {name}", operation)

        return None
        
    def verify(self):
        name = self.cfg.get("name")

        if not (create_props := self.cfg.get("create_props", None)):
            raise Exception(f"Cannot find element create_props in configuration.")

        verify_device(self.api, name, create_props)
        verify_device_communication(self.api, name)

        return None    
    
class TestCase_ModifyDevice(TestCase):
    def get_name(self):
        return "Modify Device"
        
    def run(self):
        name = self.cfg.get("name")

        if not (modify_props := self.cfg.get("modify_props", None)):
            raise Exception(f"Cannot find element modify_props in configuration.")

        def operation():        
            response = Device.update(self.api, name=name, **modify_props)

            if not response:
                raise Exception(f"Device {name} hasn't been created")
        
            #if log_level <= tutils.INFO:
            #    GetResponse(response=response).print()      

        tutils.exec_test(f"Modifying device {name}", operation)

        return None        

    def verify(self):
        name = self.cfg.get("name")

        if not (modify_props := self.cfg.get("modify_props", None)):
            raise Exception(f"Cannot find element modify_props in configuration.")
                
        verify_device(self.api, name, modify_props)
          
        return None    
    
class TestCase_DeleteDevice(TestCase):
    def get_name(self):
        return "Delete Device"
        
    def run(self):
        delete_device(self.api, self.cfg.get("name"))

    def verify(self):
        name = self.cfg.get("name")

        def operation():
            if dmo.get_resources(api_client=self.api, resourceName=name):
                raise Exception(f"Delete request didn't delete device {name}.")
        tutils.exec_test(f"Checking if device {name} has been deleted", operation)

        return None      
        
###########################################################################            
# Performs the following test cases:
# - create device
###########################################################################            
class Tests00(TestSet):
    def __init__(self, api: ApiClient):
        super().__init__(api, "Device Tests 00")
    
    def get_sample_configuration_varname(self)->str:
        return "Device01Cfg"

    def get_test_cases(self, cfg:dict)->List[TestCase]:
        return [
            TestCase_CreateDevice(api=self.api, cfg=cfg),
        ]
    
    def test_pre_cleanup(self, cfg:dict)->str:
        name = cfg.get("name")
        def operation():
            if (device := Device.get(self.api, name=name)):
                Device.delete(self.api, name=name)
            return None
        return tutils.exec_test(f"Deleting device {name}", operation)
    
    def test_post_cleanup(self, cfg:dict)->str:
        name = cfg.get("name")
        def operation():
            if (device := Device.get(self.api, name=name)):
                Device.delete(self.api, name=name)
            return None
        return tutils.exec_test(f"Deleting device {name}", operation)

class Tests00_CLI(TestCLI_Base):
    def get1LineHelp(self):
        return "Runs the device test case 00: create device."
    
    def get_test_set(self, api: ApiClient):
        return Tests00(api)

###########################################################################            
# Performs the following test cases:
# - create device
# - modify device
# - delete device
###########################################################################    
class Tests01(TestSet):
    def __init__(self, api: ApiClient):
        super().__init__(api, "Device Tests 01")
    
    def get_sample_configuration_varname(self)->str:
        return "Device01Cfg"

    def get_test_cases(self, cfg:dict)->List[TestCase]:
        return [
            TestCase_CreateDevice(api=self.api, cfg=cfg),
            TestCase_ModifyDevice(api=self.api, cfg=cfg),
            TestCase_DeleteDevice(api=self.api, cfg=cfg)
        ]
    
    def test_pre_cleanup(self, cfg:dict)->str:
        name = cfg.get("name")
        def operation():
            if (device := Device.get(self.api, name=name)):
                Device.delete(self.api, name=name)
            return None
        return tutils.exec_test(f"Deleting device {name}", operation)
    
    def test_post_cleanup(self, cfg:dict)->str:
        name = cfg.get("name")
        def operation():
            if (device := Device.get(self.api, name=name)):
                Device.delete(self.api, name=name)
            return None
        return tutils.exec_test(f"Deleting device {name}", operation)

class Tests01_CLI(TestCLI_Base):
    def get1LineHelp(self):
        return "Runs the device test case 01: create, modify, delete device."
    
    def get_test_set(self, api: ApiClient):
        return Tests01(api)
    